# -*- coding: utf-8 -*-
# @Time    : 14/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from typing import Tuple
from openpy_fxts.base_tf import tf, tkl, K, tkr, value_miss
from openpy_fxts.architectures.layers.attention_class import SeqSelfAttention


def type_rnn(name: str = None):
    if name == 'RNN':
        return tkl.SimpleRNN
    if name == 'LSTM':
        return tkl.LSTM
    if name == 'GRU':
        return tkl.GRU
    else:
        return tkl.GRU


class dense_drop(tkl.Layer):

    def __init__(
            self,
            n_units: int = None,
            list_act: str = None,
            l1=None,
            l2=None,
            use_drop:
            bool = False,
            rate: float = None
    ):
        super(dense_drop, self).__init__()
        self.is_drop = use_drop
        self.hidden_dense, self.hidden_drop = [], []
        for j, (unit, act) in enumerate(zip(n_units, list_act)):
            self.hidden_dense.append(tkl.Dense(units=unit, activation=act, kernel_regularizer=tkr.L1L2(l1=l1, l2=l2)))
            if use_drop:
                self.hidden_drop.append(tkl.Dropout(rate))

    def call(self, x):
        if self.is_drop:
            for _, (dense, drop) in enumerate(zip(self.hidden_dense, self.hidden_drop)):
                x = dense(x)
                x = drop(x)
        else:
            for _, (dense) in enumerate(self.hidden_dense):
                x = dense(x)
        return x


class multi_timedist_conv1d_flat(tkl.Layer):

    def __init__(
            self,
            filters: list = None,
            kernels: list = None,
            activations: str = None,
            paddings: str = None,
            dilation_rate: int = None,
            maxpooling_1d: bool = True,
            pool_size: int = 2,
            is_mask: bool = False

    ):
        super(multi_timedist_conv1d_flat, self).__init__()
        self.maxpooling_1d = maxpooling_1d
        self.is_mask = is_mask
        if is_mask:
            self.mask = tkl.Masking(mask_value=value_miss)
        self.hidden_timedist_conv1d = []
        for i, (f, k, a, p) in enumerate(zip(filters, kernels, activations, paddings)):
            if i == 0 and dilation_rate is None:
                self.hidden_timedist_conv1d.append(
                    tkl.TimeDistributed(tkl.Conv1D(filters=f, kernel_size=k, activation=a, padding=p, dilation_rate=2))
                )
            else:
                self.hidden_timedist_conv1d.append(
                    tkl.TimeDistributed(tkl.Conv1D(filters=f, kernel_size=k, activation=a, padding=p))
                )
        if maxpooling_1d:
            self.maxpool = tkl.TimeDistributed(tkl.MaxPooling1D(pool_size=pool_size))
        self.timedist_flatten = tkl.TimeDistributed(tkl.Flatten())

    def call(self, x):
        '''
        if self.is_mask:
            x = self.mask(x)
        '''
        for j, (layer) in enumerate(self.hidden_timedist_conv1d):
            if j == 0:
                x = layer(x)
            else:
                x = layer(x)
        if self.maxpooling_1d:
            x = self.maxpool(x)
        x = self.timedist_flatten(x)
        return x


class multi_conv1d_pool_flat_drop(tkl.Layer):

    def __init__(
            self,
            filters: list = None,
            kernels: list = None,
            activations: str = None,
            paddings: str = None,
            dilation_rate: list = None,
            use_drop: bool = False,
            rate: float = 0.3,
            maxpooling_1d: bool = True,
            pool_simple: bool = True,
            pool_global: bool = False,
            pool_size: int = 2,
            flatten: bool = True,
            is_mask: bool = False
    ):
        super(multi_conv1d_pool_flat_drop, self).__init__()
        self.is_drop = use_drop
        self.flatten = flatten
        self.maxpooling_1d = maxpooling_1d
        self.is_mask = is_mask
        if is_mask:
            self.mask = tkl.Masking(mask_value=value_miss)
        self.conv1d = []
        for i, (f, k, a, p) in enumerate(zip(filters, kernels, activations, paddings)):
            if i == 0 and dilation_rate is None:
                self.conv1d.append(tkl.Conv1D(filters=f, kernel_size=k, activation=a, padding=p, dilation_rate=2))
            else:
                self.conv1d.append(tkl.Conv1D(filters=f, kernel_size=k, activation=a, padding=p))
        if maxpooling_1d:
            if pool_simple:
                self.maxpooling_1d = tkl.MaxPooling1D(pool_size)
            elif pool_global:
                self.maxpooling_1d = tkl.GlobalAveragePooling1D(pool_size)
            else:
                self.maxpooling_1d = tkl.MaxPooling1D(pool_size)
        if flatten:
            self.flatten = tkl.Flatten()
        if use_drop:
            self.drop = tkl.Dropout(rate)


    def call(self, x):
        if self.is_mask:
            x = self.mask(x)
        for layer in self.conv1d:
            x = layer(x)
        if self.maxpooling_1d:
            x = self.maxpooling_1d(x)
        if self.flatten:
            x = self.flatten(x)
        if self.is_drop:
            x = self.drop(x)
        return x


class dense_multi_out(tkl.Layer):
    def __init__(
            self,
            n_future: int = None,
            n_out_ft: int = None,
            activation: str = None
    ):
        super(dense_multi_out, self).__init__()
        self.dense_out = tkl.Dense(n_future * n_out_ft, activation=activation)
        self.reshape = tkl.Reshape((n_future, n_out_ft))

    def call(self, x):
        x = self.dense_out(x)
        x = self.reshape(x)
        return x


class dense_reshape_conv1d_multi_out(tkl.Layer):
    def __init__(
            self,
            n_future: int = None,
            n_out_ft: int = None,
            activation: str = None,
            padding: str = 'same'
    ):
        super(dense_reshape_conv1d_multi_out, self).__init__()
        self.dense_out = tkl.Dense(n_future * n_out_ft, activation=activation)
        self.reshape = tkl.Reshape((n_future, n_out_ft))
        self.conv1d = tkl.Conv1D(n_out_ft, 1, padding=padding)

    def call(self, x):
        x = self.dense_out(x)
        x = self.reshape(x)
        x = self.conv1d(x)
        return x


class multi_rnn_drop(tkl.Layer):
    def __init__(
            self,
            mdl: object = None,
            units: list = None,
            activations: list = None,
            sequences: list = None,
            bidirectional: bool = None,
            use_drop: bool = None,
            rate: float = None,
            is_mask: bool = False
    ):
        super(multi_rnn_drop, self).__init__()
        self.is_mask = is_mask
        if is_mask:
            self.mask = tkl.Masking(mask_value=value_miss)
        self.hidden_rnn = []
        for _, (n_units, act, seq) in enumerate(zip(units, activations, sequences)):
            self.hidden_rnn.append(
                rnn_drop(
                    mdl=mdl,
                    n_units=n_units,
                    activation=act,
                    return_seq=seq,
                    bidirectional=bidirectional,
                    use_drop=use_drop,
                    rate=rate
                )
            )

    def call(self, x):
        for j, (rnn) in enumerate(self.hidden_rnn):
            if j == 0:
                if self.is_mask:
                    x = self.mask(x)
                x = rnn(x)
            else:
                x = rnn(x)
        return x


class rnn_drop(tkl.Layer):
    def __init__(
            self,
            mdl: object = None,
            n_units: int = None,
            activation: str = None,
            return_seq: bool = False,
            bidirectional: bool = False,
            use_drop: bool = False,
            rate: float = 0.2
    ):
        super(rnn_drop, self).__init__()
        self.use_drop = use_drop
        if bidirectional:
            self.rnn = tkl.Bidirectional(mdl(units=n_units, activation=activation, return_sequences=return_seq))
        else:
            self.rnn = mdl(units=n_units, activation=activation, return_sequences=return_seq)
        if self.use_drop:
            self.drop = tkl.Dropout(rate=rate)

    def call(self, x):
        x = self.rnn(x)
        if self.use_drop:
            x = self.drop(x)
        return x


class rnn_attention_conv1d(tkl.Layer):

    def __init__(
            self,
            mdl: object = None,
            bidirectional: bool = True,
            n_units: int = None,
            rnn_act: str = None,
            return_seq: str = None,
            att_type: str = None,
            kernel_reg: float = None,
            bias_reg: float = None,
            reg_weight: float = None,
            n_att: int = None,
            num_heads=None,
            key_dim=None,
            dropout=None,
            n_filters: int = None,
            kernel: int = None,
            conv_act: str = None,
            padding: str = None,
            simple_pool: bool = False,
            global_pool: bool = False,
            pool_size: int = None
    ):
        super(rnn_attention_conv1d, self).__init__()
        if bidirectional:
            self.rnn = tkl.Bidirectional(
                mdl(
                    units=n_units,
                    activation=rnn_act,
                    return_sequences=return_seq
                )
            )
        else:
            self.rnn = mdl(
                units=n_units,
                activation=rnn_act,
                return_sequences=return_seq
            )
        self.att_type = att_type
        if att_type == 'Bahdanau' or att_type == 'Luong':
            if att_type == 'Bahdanau':
                attention_type = SeqSelfAttention.ATTENTION_TYPE_ADD
            elif att_type == 'Luong':
                attention_type = SeqSelfAttention.ATTENTION_TYPE_MUL
            self.attention = SeqSelfAttention(
                attention_type=attention_type,
                kernel_regularizer=tkr.l2(kernel_reg),
                bias_regularizer=tkr.l1(bias_reg),
                attention_regularizer_weight=reg_weight,
                name=f'Attention{n_att}'
            )
        if att_type == 'MultiHeadAttention':
            self.attention = tkl.MultiHeadAttention(
                num_heads=num_heads,
                key_dim=key_dim,
                dropout=dropout
            )

        self.conv1d = tkl.Conv1D(
            filters=n_filters,
            kernel_size=kernel,
            activation=conv_act,
            padding=padding
        )
        if simple_pool:
            self.maxpool1d = tkl.MaxPool1D(pool_size)
        elif global_pool:
            self.maxpool1d = tkl.GlobalAveragePooling1D()
        else:
            self.maxpool1d = tkl.MaxPool1D()

    def call(self, x):
        x = self.rnn(x)
        if self.att_type == 'Bahdanau' or self.att_type == 'Luong':
            x = self.attention(x)
        if self.att_type == 'MultiHeadAttention':
            x = self.attention(x, x)
        x = self.conv1d(x)
        x = self.maxpool1d(x)
        return x


class multi_rnn_attention_conv1d(tkl.Layer):
    def __init__(
            self,
            mdl: object = None,
            bidirectional: bool = True,
            units: list = None,
            rnn_act: list = None,
            return_seq: list = None,
            att_type: str = None,
            kernel_reg: float = None,
            bias_reg: float = None,
            reg_weight: float = None,
            n_filters: list = None,
            kernels: list = None,
            conv_act: list = None,
            paddings: list = None,
            simple_pool: bool = None,
            global_pool: bool = None,
            pool_size: int = None
    ):
        super(multi_rnn_attention_conv1d, self).__init__()
        self.hidden_birnn_att_conv_layers = []
        for i, (unit, act_rnn, seq, ftr, knl, act_conv, pad) in enumerate(
                zip(units, rnn_act, return_seq, n_filters, kernels, conv_act, paddings)):
            self.hidden_birnn_att_conv_layers.append(
                rnn_attention_conv1d(
                    mdl=mdl,
                    bidirectional=bidirectional,
                    n_units=unit,
                    rnn_act=act_rnn,
                    return_seq=seq,
                    att_type=att_type,
                    kernel_reg=kernel_reg,
                    bias_reg=bias_reg,
                    reg_weight=reg_weight,
                    n_att=i,
                    n_filters=ftr,
                    kernel=knl,
                    conv_act=act_conv,
                    padding=pad,
                    simple_pool=simple_pool,
                    global_pool=global_pool,
                    pool_size=pool_size
                )
            )

    def call(self, x):
        for j, (birnn_att_conv) in enumerate(self.hidden_birnn_att_conv_layers):
            if j == 0:
                x = birnn_att_conv(x)
            else:
                x = birnn_att_conv(x)
        return x
        

class multi_birnn_multihead_attention_conv1d(tkl.Layer):
    def __init__(
            self,
            mdl: object = None,
            bidirectional: bool = True,
            units: list = None,
            rnn_act: list = None,
            return_seq: list = None,
            att_type: str = None,
            num_heads: int = None,
            key_dim: int = None,
            dropout: float = None,
            n_filters: list = None,
            kernels: list = None,
            conv_act: list = None,
            paddings: list = None,
            simple_pool: bool = None,
            global_pool: bool = None,
            pool_size: int = None
    ):
        super(multi_birnn_multihead_attention_conv1d, self).__init__()
        self.hidden_birnn_att_conv_layers = []
        for i, (unit, act_rnn, seq, ftr, knl, act_conv, pad) in enumerate(
                zip(units, rnn_act, return_seq, n_filters, kernels, conv_act, paddings)):
            self.hidden_birnn_att_conv_layers.append(
                rnn_attention_conv1d(
                    mdl=mdl,
                    bidirectional=bidirectional,
                    n_units=unit,
                    rnn_act=act_rnn,
                    return_seq=seq,
                    att_type=att_type,
                    num_heads=num_heads,
                    key_dim=key_dim,
                    dropout=dropout,
                    n_filters=ftr,
                    kernel=knl,
                    conv_act=act_conv,
                    padding=pad,
                    simple_pool=simple_pool,
                    global_pool=global_pool,
                    pool_size=pool_size
                )
            )

    def call(self, x):
        for j, (birnn_att_conv) in enumerate(self.hidden_birnn_att_conv_layers):
            if j == 0:
                x = birnn_att_conv(x)
            else:
                x = birnn_att_conv(x)
        return x


class Time2Vec(tkl.Layer):

    def __init__(self, output_dim=None, **kwargs):
        self.output_dim = output_dim
        super(Time2Vec, self).__init__(**kwargs)

    def build(self, input_shape):
        self.W = self.add_weight(
            name='W',
            shape=(input_shape[-1], self.output_dim),
            initializer='uniform',
            trainable=True)

        self.P = self.add_weight(
            name='P',
            shape=(input_shape[1], self.output_dim),
            initializer='uniform',
            trainable=True
        )
        self.w = self.add_weight(
            name='w',
            shape=(input_shape[1], 1),
            initializer='uniform',
            trainable=True
        )
        self.p = self.add_weight(
            name='p',
            shape=(input_shape[1], 1),
            initializer='uniform',
            trainable=True
        )
        super(Time2Vec, self).build(input_shape)

    def call(self, x):
        original = self.w * x + self.p
        sin_trans = K.sin(K.dot(x, self.W) + self.P)
        return K.concatenate([sin_trans, original], -1)


class time2vec_birnn_layer(tkl.Layer):

    def __init__(
            self,
            mdl: object = None,
            output_dim: int = None,
            unit: int = None,
            rnn_act: int = None,
            rnn_seq: int = None,
            is_drop: bool = None,
            rate: float = None,
    ):
        super(time2vec_birnn_layer, self).__init__()
        self.is_drop = is_drop
        self.time2vec = Time2Vec(output_dim=output_dim)
        self.birnn = tkl.Bidirectional(
            mdl(
                units=unit,
                activation=rnn_act,
                return_sequences=rnn_seq
            )
        )
        if is_drop:
            self.drop = tkl.Dropout(rate)

    def call(self, x):
        x = self.time2vec(x)
        x = self.birnn(x)
        if self.is_drop:
            x = self.drop(x)
        return x


class multi_time2vec_birnn(tkl.Layer):

    def __init__(
            self,
            mdl: object = None,
            output_dims: list = None,
            units: list = None,
            rnn_acts: list = None,
            return_seq: list = None,
            is_drop: bool = None,
            rate: float = None,
    ):
        super(multi_time2vec_birnn, self).__init__()
        self.hidden_time2vec = []
        for i, (dim, unit, act_rnn, seq,) in enumerate(zip(output_dims, units, rnn_acts, return_seq)):
            self.hidden_time2vec.append(
                time2vec_birnn_layer(
                    mdl=mdl,
                    output_dim=dim,
                    unit=unit,
                    rnn_act=act_rnn,
                    rnn_seq=seq,
                    is_drop=is_drop,
                    rate=rate
                )
            )

    def call(self, x):
        for j, (layer) in enumerate(self.hidden_time2vec):
            if j == 0:
                x = layer(x)
            else:
                x = layer(x)
        return x




class InnerConv1DBlock(tkl.Layer):

    def __init__(self, filters: int, h: float, kernel_size: int, neg_slope: float = .01, dropout: float = .5, **kwargs):
        if filters <= 0 or h <= 0:
            raise ValueError('filters and h must be positive')
        super(InnerConv1DBlock, self).__init__(**kwargs)
        self.conv1d = tkl.Conv1D(max(round(h * filters), 1), kernel_size, padding='same')
        self.leakyrelu = tkl.LeakyReLU(neg_slope)
        self.dropout = tkl.Dropout(dropout)
        self.conv1d2 = tkl.Conv1D(filters, kernel_size, padding='same')
        self.tanh = tf.keras.activations.tanh

    def call(self, input_tensor, training=None):
        x = self.conv1d(input_tensor)
        x = self.leakyrelu(x)

        if training:
            x = self.dropout(x)

        x = self.conv1d2(x)
        x = self.tanh(x)
        return x


class SciBlock(tkl.Layer):
    def __init__(self, features: int, kernel_size: int, h: int, **kwargs):
        """
        :param features: number of features in the output
        :param kernel_size: kernel size of the convolutional layers
        :param h: scaling factor for convolutional module
        """

        super(SciBlock, self).__init__(**kwargs)
        self.features = features
        self.kernel_size = kernel_size
        self.h = h

    def build(self, input_shape):
        self.conv1ds = {
            k: InnerConv1DBlock(
                filters=self.features,
                h=self.h,
                kernel_size=self.kernel_size,
                name=k
            ) for k in ['psi', 'phi', 'eta', 'rho']
        }  # regularize?
        super().build(input_shape)
        # [layer.build(input_shape) for layer in self.conv1ds.values()]  # unneeded?

    def call(self, inputs, training=None):
        F_odd, F_even = inputs[:, ::2], inputs[:, 1::2]

        # Interactive learning as described in the paper
        F_s_odd = F_odd * tf.math.exp(self.conv1ds['phi'](F_even))
        F_s_even = F_even * tf.math.exp(self.conv1ds['psi'](F_odd))

        F_prime_odd = F_s_odd + self.conv1ds['rho'](F_s_even)
        F_prime_even = F_s_even - self.conv1ds['eta'](F_s_odd)

        return F_prime_odd, F_prime_even


class Interleave(tkl.Layer):
    """A layer used to reverse the even-odd split operation."""

    def __init__(self, **kwargs):
        super(Interleave, self).__init__(**kwargs)

    def interleave(self, slices):
        if not slices:
            return slices
        elif len(slices) == 1:
            return slices[0]

        mid = len(slices) // 2
        even = self.interleave(slices[:mid])
        odd = self.interleave(slices[mid:])

        shape = tf.shape(even)
        return tf.reshape(tf.stack([even, odd], axis=3), (shape[0], shape[1] * 2, shape[2]))

    def call(self, inputs):
        return self.interleave(inputs)


class SciNet(tkl.Layer):
    def __init__(
            self,
            horizon: int,
            features: int,
            levels: int,
            h: int,
            kernel_size: int,
            regularizer: Tuple[float, float] = (0, 0),
            **kwargs
    ):
        """
        :param horizon: number of time stamps in output
        :param features: number of features in output
        :param levels: height of the binary tree + 1
        :param h: scaling factor for convolutional module in each SciBlock
        :param kernel_size: kernel size of convolutional module in each SciBlock
        :param regularizer: activity regularization (not implemented)
        """

        if levels < 1:
            raise ValueError('Must have at least 1 level')
        super(SciNet, self).__init__(**kwargs)
        self.horizon = horizon
        self.features = features
        self.levels = levels
        self.interleave = Interleave()
        self.flatten = tkl.Flatten()
        self.dense = tkl.Dense(
            horizon * features,
            kernel_regularizer=tf.keras.regularizers.L1L2(0.001, 0.01),
            # activity_regularizer=L1L2(0.001, 0.01)
        )
        # self.regularizer = tkl.ActivityRegularization(l1=regularizer[0], l2=regularizer[1])

        # tree of sciblocks
        self.sciblocks = [SciBlock(features=features, kernel_size=kernel_size, h=h) for _ in range(2 ** levels - 1)]

    def build(self, input_shape):
        if input_shape[1] / 2 ** self.levels % 1 != 0:
            raise ValueError(
                f'timestamps {input_shape[1]} must be evenly divisible by a tree with {self.levels} levels'
            )
        super().build(input_shape)
        # [layer.build(input_shape) for layer in self.sciblocks]  # input_shape

    def call(self, inputs, training=None):
        # cascade input down a binary tree of sci-blocks
        lvl_inputs = [inputs]  # inputs for current level of the tree
        for i in range(self.levels):
            i_end = 2 ** (i + 1) - 1
            i_start = i_end - 2 ** i
            lvl_outputs = [
                output for j, tensor in zip(range(i_start, i_end), lvl_inputs) for output in self.sciblocks[j](tensor)
            ]
            lvl_inputs = lvl_outputs

        x = self.interleave(lvl_outputs)
        x += inputs

        # not sure if this is the correct way of doing it. The paper merely said to use a fully connected layer to
        # produce an output. Can't use TimeDistributed wrapper. It would force the layer's timestamps to match that of
        # the input -- something SCINet is supposed to solve
        x = self.flatten(x)
        x = self.dense(x)
        x = tf.reshape(x, (-1, self.horizon, self.features))

        return x


class StackedSciNet(tkl.Layer):
    def __init__(
            self,
            horizon: int,
            features: int,
            stacks: int,
            levels: int,
            h: int,
            kernel_size: int,
            regularizer: Tuple[float, float] = (0, 0),
            **kwargs
    ):
        """
        :param horizon: number of time stamps in output
        :param stacks: number of stacked SciNets
        :param levels: number of levels for each SciNet
        :param h: scaling factor for convolutional module in each SciBlock
        :param kernel_size: kernel size of convolutional module in each SciBlock
        :param regularizer: activity regularization (not implemented)
        """

        if stacks < 1:
            raise ValueError('Must have at least 1 stack')
        super(StackedSciNet, self).__init__(**kwargs)
        self.horizon = horizon
        self.scinets = [SciNet(
            horizon=horizon,
            features=features,
            levels=levels,
            h=h,
            kernel_size=kernel_size,
            regularizer=regularizer) for _ in range(stacks)
        ]
        # self.mse_fn = tf.keras.metrics.MeanSquaredError()
        # self.mae_fn = tf.keras.metrics.MeanAbsoluteError()

    # def build(self, input_shape):
    #     super().build(input_shape)
    #     [stack.build(input_shape) for stack in self.scinets]

    def call(self, inputs, targets=None, sample_weights=None, training=None):
        outputs = []
        for scinet in self.scinets:
            x = scinet(inputs)
            outputs.append(x)  # keep each stack's output for intermediate supervision
            inputs = tf.concat([x, inputs[:, x.shape[1]:, :]], axis=1)

        if targets is not None:
            # Calculate metrics
            # mse = self.mse_fn(targets, x, sample_weights)
            # mae = self.mae_fn(targets, x, sample_weights)
            # self.add_metric(mse, name='mean_squared_error')
            # self.add_metric(mae, name='mean_absolute_error')

            if training:
                # Calculate loss as sum of mean of norms of differences between output and input feature vectors for
                # each stack
                stacked_outputs = tf.stack(outputs)
                differences = stacked_outputs - targets
                loss = tf.linalg.normalize(differences, axis=1)[1]
                loss = tf.reshape(loss, (-1, self.horizon))
                loss = tf.reduce_sum(loss, 1)
                loss = loss / self.horizon
                loss = tf.reduce_sum(loss)
                self.add_loss(loss)

        return x

    def get_config(self):
        return super().get_config()





# -*- coding: utf-8 -*-
# @Time    : 05/09/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : removal_data.py
# @Software: PyCharm

from statsmodels.tsa.seasonal import seasonal_decompose
import numpy as np


class seasonal:

    @staticmethod
    def nullify_outliers(col, period: int = None):
        """
        Remove outliers using seasonal decomposition
        :param col: (pd.Series) df to remove outliers for
        :return: (pd.Series) df with linearly interpolated values
        """
        #n_nan = col.isna().sum()
        decomp = seasonal_decompose(col, model='additive', period=period)
        resid = np.array(decomp.resid)

        # na fill outliers
        per_25, per_75 = np.percentile(resid[~np.isnan(resid)], [25, 75])
        upper = per_75 + (per_75 - per_25) * 1.5
        lower = per_25 - (per_75 - per_25) * 1.5
        col.iloc[np.argwhere((resid > upper) | (resid < lower)).flatten()] = np.nan

        return col





# -*- coding: utf-8 -*-
# @Time    : 04/09/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : utils_plot.py
# @Software: PyCharm

import seaborn as sns
import matplotlib.pyplot as plt


class plot_scenarios:
    
    def __init__(self, data):
        self.data = data

    def lineplot_SingleUser(self):
        self.data = self.data.reset_index()
        self.data[self.data.columns[0]] = self.data[self.data.columns[0]].apply(lambda x: x.strftime('%H:%M'))
        self.data.set_index(self.data.columns[0], inplace=True)
        g = sns.lineplot(x=self.data.index, y=self.data.columns[0], hue='date', data=self.data)
        g.set_xticklabels(g.get_xticklabels(), rotation=90)
        plt.show()

    def boxplot_SingleUser(self):
        self.data = self.data.reset_index()
        self.data[self.data.columns[0]] = self.data[self.data.columns[0]].apply(lambda x: x.strftime('%H:%M'))
        self.data.set_index(self.data.columns[0], inplace=True)
        g = sns.boxplot(x=self.data.index, y=self.data.columns[0], data=self.data)
        g.set_xticklabels(g.get_xticklabels(), rotation=90)
        plt.show()



from openpy_fxts.models.imputation.init_modules import hpc_dataframe
from openpy_fxts.models.imputation.utilities.Utils import Utils, Metrics, Preprocessing, ft_eng
from openpy_fxts.models.imputation.generate_missing_data.MCAR_MAR_MNAR_Pattern import generate_missing_data
from openpy_fxts.models.imputation.Iterative import iterative

# Models Simple
from openpy_fxts.models.imputation.simple import imp_basic
from openpy_fxts.models.imputation.mdls_ml.models import imp_ml

# MLP
# Simple
from openpy_fxts.models.imputation.mdls_dl.simple.dlm_MLP import MLP_Dense_class
# Conv1D
# Simple
from openpy_fxts.models.imputation.mdls_dl.simple.dlm_Conv1D import Conv1D_Dense_class, Multi_Conv1D_Dense_class
# Hybrids
from openpy_fxts.models.imputation.mdls_dl.hybrids.dlm_Conv1D import Conv1D_BiRNN_class, TimeDist_Conv1D_BiRNN_class
from openpy_fxts.models.imputation.mdls_dl.hybrids.dlm_Conv1D import Conv1D_BiRNN_Attention_class
# BiRNN
# Simple
from openpy_fxts.models.imputation.mdls_dl.simple.dlm_BiRNN import BiRNN_Dense_class, Multi_BiRNN_Dense_class
# Hybrids
from openpy_fxts.models.imputation.mdls_dl.hybrids.dlm_BiRNN import BiRNN_Conv1D_class, BiRNN_TimeDist_Dense_class
from openpy_fxts.models.imputation.mdls_dl.hybrids.dlm_BiRNN import BiRNN_Luong_Attention_Conv1D_class
from openpy_fxts.models.imputation.mdls_dl.hybrids.dlm_BiRNN import BiRNN_Bahdanau_Attention_Conv1D_class
from openpy_fxts.models.imputation.mdls_dl.hybrids.dlm_BiRNN import BiRNN_MultiHeadAttention_Conv1D_class
# Model: Others with BiRNN
from openpy_fxts.models.imputation.mdls_dl.hybrids.dlm_Others import TCN_BiRNN_class, Time2Vec_BiRNN_class
# Model: Encoder Decoder
from openpy_fxts.models.imputation.mdls_dl.hybrids.dlm_EncDec import EncDec_BiRNN_class, EncDec_Conv1D_BiRNN_class
# Test module
from openpy_fxts.test_lib.run_dl_mdls_mdits import run_all_models

from .data_visualization.Visualization import utils_view
view = utils_view()







from openpy_fxts.base_tf import tkm, tkl
from openpy_fxts.architectures.layers.layers_class import dense_multi_out, rnn_drop, type_rnn
from openpy_fxts.architectures.layers.layers_class import multi_rnn_drop, multi_conv1d_pool_flat_drop, multi_timedist_conv1d_flat
from openpy_fxts.architectures.layers.attention_class import attention
from openpy_fxts.architectures.fxts.arch_dicts import _get_dict_fxts


class conv1d_birnn_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(conv1d_birnn_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('Conv1D_BiRNN').config_conv1d()
        self.n_future = n_future,
        self.n_out_ft = n_out_ft,
        self.config = config
        self.multi_hidden_conv1d = multi_conv1d_pool_flat_drop(
            filters=config['conv1d_layers']['filters'],
            kernels=config['conv1d_layers']['kernels'],
            activations=config['conv1d_layers']['activations'],
            paddings=config['conv1d_layers']['paddings'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
            maxpooling_1d=True,
            pool_size=config['conv1d_layers']['pool_size'],
            flatten=False
        )
        self.multi_hidden_rnn = multi_rnn_drop(
            mdl=type_rnn(config['rnn_layers']['type']),
            units=config['rnn_layers']['units'],
            activations=config['rnn_layers']['activations'],
            sequences=config['rnn_layers']['sequences'],
            bidirectional=config['rnn_layers']['bidirectional'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate']
        )
        if config['dense_layer']['activation']:
            self.dense_layer = tkl.Dense(
                units=config['dense_layer']['units'],
                activation=config['dense_layer']['activation']
            )
            self.drop = tkl.Dropout(rate=config['dropout']['rate'])
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.multi_hidden_conv1d(inputs)
        x = self.multi_hidden_rnn(x)
        if self.config['dense_layer']['activation']:
            x = self.dense_layer(x)
            x = self.drop(x)
        x = self.dense_out(x)
        return x


class timedist_conv1d_birnn_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(timedist_conv1d_birnn_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('TimeDist_Conv1D_BiRNN').config_conv1d()
        self.n_future = n_future,
        self.n_out_ft = n_out_ft,
        self.config = config
        self.multi_hidden_conv1d = multi_timedist_conv1d_flat(
            filters=config['conv1d_layers']['filters'],
            kernels=config['conv1d_layers']['kernels'],
            activations=config['conv1d_layers']['activations'],
            paddings=config['conv1d_layers']['paddings'],
            maxpooling_1d=config['conv1d_layers']['maxpooling'],
            pool_size=config['conv1d_layers']['pool_size']
        )
        self.rnn_layer = rnn_drop(
            mdl=type_rnn(config['rnn_layer']['type']),
            n_units=config['rnn_layer']['units'],
            activation=config['rnn_layer']['activation'],
            return_seq=config['rnn_layer']['sequences'],
            bidirectional=config['rnn_layer']['bidirectional']
        )
        if config['dropout']['activate']:
            self.drop = tkl.Dropout(rate=config['dropout']['rate'])
        self.dense_layer = tkl.Dense(
            units=config['dense_layer']['units'],
            activation=config['dense_layer']['activation']
        )
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.multi_hidden_conv1d(inputs)
        x = self.rnn_layer(x)
        if self.config['dropout']['activate']:
            x = self.drop(x)
        x = self.dense_layer(x)
        x = self.dense_out(x)
        return x


class conv1d_birnn_attention_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(conv1d_birnn_attention_dense, self).__init__()
        if config is None:
            config = _get_dict_fxts('Conv1D_BiRNN_Attention').config_conv1d()
        self.n_future = n_future,
        self.n_out_ft = n_out_ft,
        self.config = config
        self.hidden_conv1d = multi_conv1d_pool_flat_drop(
            filters=config['conv1d_layers']['filters'],
            kernels=config['conv1d_layers']['kernels'],
            activations=config['conv1d_layers']['activations'],
            paddings=config['conv1d_layers']['paddings'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'],
            maxpooling_1d=True,
            pool_size=config['conv1d_layers']['pool_size'],
            flatten=False
        )
        self.multi_hidden_birnn = multi_rnn_drop(
            mdl=type_rnn(config['rnn_layers']['type']),
            units=config['rnn_layers']['units'],
            activations=config['rnn_layers']['activations'],
            sequences=config['rnn_layers']['sequences'],
            bidirectional=config['rnn_layers']['bidirectional'],
            use_drop=config['dropout']['activate'],
            rate=config['dropout']['rate'])
        self.attention = attention()
        self.flatten = tkl.Flatten()
        if config['dense_layer']['activation']:
            self.dense_layer = tkl.Dense(
                units=config['dense_layer']['units'],
                activation=config['dense_layer']['activation']
            )
            self.drop = tkl.Dropout(rate=config['dropout']['rate'])
        self.dense_out = dense_multi_out(
            n_future,
            n_out_ft,
            config['output_layer']['activation']
        )

    def call(self, inputs, training=True, **kwargs):
        x = self.hidden_conv1d(inputs)
        x = self.multi_hidden_birnn(x)
        x = self.attention(x)
        x = self.flatten(x)
        if self.config['dense_layer']['activation']:
            x = self.dense_layer(x)
            x = self.drop(x)
        x = self.dense_out(x)
        return x


# -*- coding: utf-8 -*-
# @Time    : 24/03/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import logging
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from ..utilities.Utils import Utils

log_py = logging.getLogger(__name__)


class utils_view:

    def density_true_missing(
            self,
            df_miss: pd.DataFrame = None,
            dict_mdls: dict = None,
            view: bool = False,
            cumulative=False
    ):

        col_miss = Utils(df_miss).list_column_missing()
        list_x = []
        list_y = []
        aux = 0
        for i in range(0, len(col_miss), 2):
            list_x.append(aux)
            list_x.append(aux)
            list_y.append(0)
            list_y.append(1)
            aux += 1

        sns.set(style='darkgrid')
        fig, ax = plt.subplots(figsize=(29, 12), nrows=int(round(len(col_miss) / 2, 1)), ncols=2)
        for col, i, j in zip(col_miss, list_x, list_y):
            if len(col_miss) <= 2:
                for key, value in dict_mdls.items():
                    sns.kdeplot(
                        x=value[col][df_miss.isnull().any(axis=1)], label=key, ax=ax[j], cumulative=cumulative
                    )
                ax[j].legend()
            else:
                for key, value in dict_mdls.items():
                    sns.kdeplot(
                        x=value[col][df_miss.isnull().any(axis=1)], label=key, ax=ax[i][j], cumulative=cumulative
                    )
                ax[i][j].legend()
        if view:
            plt.show()

    def data_correlation_heatmap(self, df):
        sns.heatmap(
            data=df.corr(),
            cmap=sns.diverging_palette(20, 230, as_cmap=True),
            center=0,
            vmin=-1,
            vmax=1,
            linewidths=0.5,
            annot=True
        )



# -*- coding: utf-8 -*-
# @Time    : 06/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import os
import numpy as np
import tensorflow as tf
import random as python_random

tkm = tf.keras.models
tkl = tf.keras.layers
tkloss = tf.keras.losses
tko = tf.keras.optimizers
tku = tf.keras.utils
K = tf.keras.backend
tkr = tf.keras.regularizers
tfk = tf.keras
tkc = tf.keras.callbacks
tkmtc = tf.keras.metrics

value_miss = 0
seed_lab = 44


seed_value = 1200
os.environ["PYTHONHASHSEED"] = str(seed_value)


def reset_seeds(seed_value):
    np.random.seed(seed_value)
    python_random.seed(seed_value)
    tf.random.set_seed(seed_value)
    tf.keras.utils.set_random_seed(seed_value)



# -*- coding: utf-8 -*-
# @Time    : 24/03/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import logging

import pandas as pd

from openpy_fxts.datasets.test_data.load_examples import _sample_test
from .generate_missing_data.MCAR_MAR_MNAR_Pattern import generate_missing_data

log_py = logging.getLogger(__name__)


def hpc_dataframe(
        data_name: str = 'HPC',
        date_init: str = '2008-11-26 00:00:00',
        date_end: str = None
):
    ex = _sample_test()
    data = ex._example(data_name, date_init, date_end)
    return data


class dataset_examples:

    def generate_missing_data(
            self,
            X: pd.DataFrame = None,
            p_miss: float = 0.40,
            mecha: str = "MCAR",
            opt=None,
            p_obs=None,
            q=None,
            seed: int = 0
    ):
        """
        Generate missing values for specifics missing-datasets mechanism and proportion of missing values.

        Parameters
        ----------
        X : torch.DoubleTensor or np.ndarray, shape (n, d)
            Data for which missing values will be simulated.
            If a numpy array is provided, it will be converted to a pytorch tensor.
        p_miss : float
            Proportion of missing values to generate for variables which will have missing values.
        mecha : str,
                Indicates the missing-datasets mechanism to be used. "MCAR" by default, "MAR", "MNAR" or "MNARsmask"
        opt: str,
                For mecha = "MNAR", it indicates how the missing-datasets mechanism is generated: using a logistic regression ("logistic"), quantile censorship ("quantile") or logistic regression for generating a self-masked MNAR mechanism ("selfmasked").
        p_obs : float
                If mecha = "MAR", or mecha = "MNAR" with opt = "logistic" or "quanti", proportion of variables with *no* missing values that will be used for the logistic masking model.
        q : float
            If mecha = "MNAR" and opt = "quanti", quantile level at which the cuts should occur.

        Returns
        ----------
        A dictionnary containing:
        'X_init': the initial datasets matrix.
        'X_incomp': the datasets with the generated missing values.
        'mask': a matrix indexing the generated missing values.s
        """
        miss_data = generate_missing_data(seed)
        df_data_miss = miss_data.produce_NaN(X.to_numpy(), p_miss, mecha, opt, p_obs, q)
        df_data_miss = pd.DataFrame(df_data_miss['X_incomp'], columns=X.columns)
        df_data_miss['datetime'] = X.index
        df_data_miss.set_index('datetime', inplace=True)

        return df_data_miss

    def data_preprocessing(self):
        pass

    def visualization_missing_data(self):
        pass

    def Simple_Imputation(self):
        pass

    def ML_Imputation(self):
        pass

    def DL_Imputation(self):
        pass

    def Forescating_Imputation(self):
        pass



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from tcn import TCN
from openpy_fxts.base_tf import tkm, tkl, value_miss
from openpy_fxts.architectures.layers.layers_class import Time2Vec


class tcn_bilstm(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(tcn_bilstm, self).__init__()
        if config is None:
            config = {
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.tcn = TCN(
            nb_filters=32,
            kernel_size=16,
            nb_stacks=4,
            dilations=(2, 4, 8, 16, 32, 64),
            padding='causal',
            use_skip_connections=True,
            dropout_rate=0.0,
            activation='tanh',
            kernel_initializer='glorot_uniform',
            use_batch_norm=False,
            use_layer_norm=False,
            use_weight_norm=False,
            return_sequences=True
        )
        self.bilstm = tkl.Bidirectional(
            tkl.LSTM(
                units=n_out_ft,
                return_sequences=False,
                activation='tanh'
            )
        )
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.tcn(x)
        x = self.bilstm(x)
        x = self.dense1(x)
        x = self.dense2(x)
        return x


class time2vec_bilstm(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(time2vec_bilstm, self).__init__()
        if config is None:
            config = {
                'time2vec': {
                    'units': 256
                },
                'dense': {
                    'activation': 'linear'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.time2vec = Time2Vec(
            config['time2vec']['units']
        )
        self.bilstm = tkl.Bidirectional(
            tkl.LSTM(
                units=n_out_ft,
                return_sequences=False,
                activation='tanh'
            )
        )
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.time2vec(x)
        x = self.bilstm(x)
        x = self.dense1(x)
        x = self.dense2(x)
        return x


import numpy as np
import pandas as pd
from sklearn.preprocessing import FunctionTransformer
from openpy_fxts.base_tf import value_miss

class complete_missing:
    pass


def sin_transformer(period):
    return FunctionTransformer(lambda x: np.sin(x / period * 2 * np.pi))


def cos_transformer(period):
    return FunctionTransformer(lambda x: np.cos(x / period * 2 * np.pi))


def _filter_y_features(df_y, y_features, n_future):
    list_filter = list()
    for i in range(len(y_features)):
        a = [x for x in list(df_y.columns) if y_features[i] in x]
        list_filter.extend(a)
    df_y = df_y[list_filter]

    dict_aux = {}
    for i in range(len(y_features)):
        a = [x for x in list(df_y.columns) if y_features[i] in x]
        dict_aux[i] = a

    list_aux = []
    for i in range(n_future):
        for key, value in dict_aux.items():
            dict_aux[key] = value
            list_aux.append(value[i])

    return df_y[list_aux]


def _series_to_supervised(
        data,
        dropnan=None,
        n_in=1,
        n_out=1,
        feat_str_at_end=True,
        feat_lag_str='IP',
        feat_lead_str='OP',
        value_replace=value_miss
):

    if data.isnull().sum().sum() > 0:
        data.fillna(value_replace, inplace=True)

    if feat_str_at_end:
        n_vars = 1 if type(data) is list else data.shape[1]
        df = pd.DataFrame(data)
        cols, names = list(), list()
        # input sequence (t-n, ... t-1)
        for i in range(n_in, 0, -1):
            cols.append(df.shift(i))
            if i < 10:
                name_i = '0' + str(i)
            else:
                name_i = str(i)
            names += [
                (
                        str(
                            pd.DataFrame(df.iloc[:, j]).columns.values
                        ).replace("']", '').replace("['", '') + '_' + feat_lag_str + '_' + name_i
                ) for j in range(n_vars)
            ]
        # forecast sequence (t, t+1, ... t+n)
        for i in range(0, n_out):
            cols.append(df.shift(-i))
            if i == 0:
                names += [
                    (
                        str(
                            pd.DataFrame(df.iloc[:, j]).columns.values
                        ).replace("']", '').replace("['", '')
                    ) for j in range(n_vars)
                ]
            else:
                if i < 10:
                    name_i = '0' + str(i)
                else:
                    name_i = str(i)
                names += [
                    (
                            str(
                                pd.DataFrame(df.iloc[:, j]).columns.values
                            ).replace("']", '').replace("['", '') + '_' + feat_lead_str + '_' + name_i
                    ) for j in range(n_vars)
                ]
        # put it all together
        agg = pd.concat(cols, axis=1)
        agg.columns = names
        # drop rows with NaN values
        if dropnan:
            agg.dropna(inplace=True)

    else:
        n_vars = 1 if type(data) is list else data.shape[1]
        df = pd.DataFrame(data)
        cols, names = list(), list()
        # input sequence (t-n, ... t-1)
        for i in range(n_in, 0, -1):
            cols.append(df.shift(i))
            names += [
                (
                    feat_lag_str + '%d' % (i) + str(
                        pd.DataFrame(df.iloc[:, j]).columns.values
                    ).replace("']", '').replace("['", '')
                ) for j in range(n_vars)
            ]
        # forecast sequence (t, t+1, ... t+n)
        for i in range(0, n_out):
            cols.append(df.shift(-i))
            if i == 0:
                names += [
                    (
                        str(pd.DataFrame(df.iloc[:, j]).columns.values).replace("']", '').replace("['", '')
                    ) for j in range(n_vars)
                ]
            else:
                if i < 10:
                    name_i = '0' + str(i)
                else:
                    name_i = str(i)
                names += [
                    (
                            str(
                                pd.DataFrame(df.iloc[:, j]).columns.values
                            ).replace("']", '').replace("['", '') + '_' + feat_lead_str + '_' + name_i
                    ) for j in range(n_vars)
                ]
        # put it all together
        agg = pd.concat(cols, axis=1)
        agg.columns = names
        # drop rows with NaN values
        if dropnan:
            agg.dropna(inplace=True)
    return agg


def _scaler_data(df_data, MinMax: bool = True, Standard: bool = None):
    if Standard:
        MinMax = False
        from sklearn.preprocessing import StandardScaler
        scaler = StandardScaler()
    if MinMax:
        from sklearn.preprocessing import MinMaxScaler
        scaler = MinMaxScaler(feature_range=(-1, 1))
    df = pd.DataFrame(
        columns=df_data.columns,
        index=df_data.index)
    df_scaler = df_data

    scalers = {}
    for i in df_data.columns:
        scaler = MinMaxScaler(feature_range=(-1, 1))
        s_s = scaler.fit_transform(df_scaler[i].values.reshape(-1, 1))
        s_s = np.reshape(s_s, len(s_s))
        scalers['scaler_' + i] = scaler
        # df.loc[:, i] = np.asarray(s_s, dtype=float)
        df_scaler[i] = s_s

    return df_scaler, scalers



# Simple models_fxts
# MLP
from openpy_fxts.models.forecast.simple.dlm_MLP import MLP_Dense_class
# Conv1D
from openpy_fxts.models.forecast.simple.dlm_Conv1D import Conv1D_Dense_class, Multi_Conv1D_Dense_class
# Model Recurrent Neural Networks (BiRNN)
from openpy_fxts.models.forecast.simple.dlm_BiRNN import BiRNN_Dense_class, Multi_BiRNN_Dense_class

# Hybrids models_fxts
# Conv1D
from openpy_fxts.models.forecast.hybrids.dlm_Conv1D import Conv1D_BiRNN_class, TimeDist_Conv1D_BiRNN_class
from openpy_fxts.models.forecast.hybrids.dlm_Conv1D import Conv1D_BiRNN_Attention_class
# BiRNN
from openpy_fxts.models.forecast.hybrids.dlm_BiRNN import BiRNN_Conv1D_class, BiRNN_TimeDist_Dense_class
from openpy_fxts.models.forecast.hybrids.dlm_BiRNN import BiRNN_Luong_Attention_Conv1D_class
from openpy_fxts.models.forecast.hybrids.dlm_BiRNN import BiRNN_Bahdanau_Attention_Conv1D_class
from openpy_fxts.models.forecast.hybrids.dlm_BiRNN import BiRNN_MultiHeadAttention_Conv1D_class
# Model: Others with BiRNN
from openpy_fxts.models.forecast.hybrids.dlm_Others import TCN_BiRNN_class, Time2Vec_BiRNN_class
# Model: Encoder Decoder
from openpy_fxts.models.forecast.hybrids.dlm_EncDec import EncDec_BiRNN_class, EncDec_Conv1D_BiRNN_class
# Stacked

# Seq2Seq


from openpy_fxts.utils import date_init_final
from openpy_fxts.datasets.import_data import hpc_dataframe


# -*- coding: utf-8 -*-
# @Time    : 27/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import pandas as pd


def date_init_final(config: dict = None):
    init = config['dataset'].index[0]
    final = config['dataset'].index[-1]
    date1 = f"{init.year}{init.month}{init.day}"
    date2 = f"{final.year}{final.month}{final.day}"
    n_in_out = f"int{config['n_past']}_out{config['n_future']}"
    fts = ""
    for i in config['y_colname']:
        fts += f"_{i}"
    name_folder = date1 + '_' + date2 + fts + '_' + n_in_out
    mttr_upd = 'mttr_' + config['MTTR'] + '_upd_' + config['Act']

    return name_folder, mttr_upd


def get_numeric_categorical(dataset: pd.DataFrame = None):
    col_type = dataset.dtypes
    numeric = col_type[col_type != object].index
    categorical = col_type[col_type == object].index
    return numeric, categorical



import tensorflow as tf

tkm = tf.keras.models
tkl = tf.keras.layers
tko = tf.keras.optimizers


def model_E1D1(n_in, n_out, n_features, units: int = 100):
    # E1D1
    # n_features ==> no of features at each timestep in the datasets.
    #
    encoder_inputs = tkl.Input(shape=(n_in, n_features))
    encoder_l1 = tkl.LSTM(units, return_state=True)
    encoder_outputs1 = encoder_l1(encoder_inputs)
    encoder_states1 = encoder_outputs1[1:]
    #
    decoder_inputs = tkl.RepeatVector(n_out)(encoder_outputs1[0])
    #
    decoder_l1 = tkl.LSTM(units, return_sequences=True)(decoder_inputs, initial_state=encoder_states1)
    decoder_outputs1 = tkl.TimeDistributed(tkl.Dense(n_features))(decoder_l1)
    #
    model_e1d1 = tf.keras.models.Model(encoder_inputs, decoder_outputs1)
    return model_e1d1


def model_E2D2(n_in, n_out, n_features, units: int = 100):
    # E2D2
    # n_features ==> no of features at each timestep in the datasets.
    #
    encoder_inputs = tkl.Input(shape=(n_in, n_features))
    encoder_l1 = tkl.LSTM(units, return_sequences=True, return_state=True)
    encoder_outputs1 = encoder_l1(encoder_inputs)
    encoder_states1 = encoder_outputs1[1:]
    encoder_l2 = tkl.LSTM(units, return_state=True)
    encoder_outputs2 = encoder_l2(encoder_outputs1[0])
    encoder_states2 = encoder_outputs2[1:]
    #
    decoder_inputs = tf.keras.layers.RepeatVector(n_out)(encoder_outputs2[0])
    #
    decoder_l1 = tkl.LSTM(units, return_sequences=True)(decoder_inputs, initial_state=encoder_states1)
    decoder_l2 = tkl.LSTM(units, return_sequences=True)(decoder_l1, initial_state=encoder_states2)
    decoder_outputs2 = tkl.TimeDistributed(tkl.Dense(n_features))(decoder_l2)
    #
    model_e2d2 = tf.keras.models.Model(encoder_inputs, decoder_outputs2)
    return model_e2d2


def model_2LSTM_2Drop_Dense(
        n_timesteps: int = None,
        n_features: int = None,
        n_outputs: int = None,
        units: int = 200,
        dropout: float = 0.2,
        optimizer=tko.RMSprop(learning_rate=0.01),
        # optimizer=tko.Adam(learning_rate=0.01),
        summary: bool = True
):
    """
    Model LSTM-Dropout-LSTM-Dropout-Dense-Reshape

    :param n_timesteps:
    :param n_features:
    :param n_outputs:
    :param units:
    :param dropout:
    :param optimizer:
    :param summary:
    :return:
    """
    # Define model.
    name_model = 'LSTM2_Drop2_Dense'
    model = tkm.Sequential()
    model.add(tkl.LSTM(units, input_shape=(n_timesteps, n_features), return_sequences=True))
    model.add(tkl.Dropout(dropout))
    model.add(tkl.LSTM(units, activation='relu'))
    model.add(tkl.Dropout(dropout))
    model.add(tkl.Dense((n_outputs * n_features), activation='linear'))
    model.add(tkl.Reshape((n_outputs, n_features)))
    # model.compile(loss='mse', optimizer='adam')
    model.compile(optimizer=optimizer, loss='mse')
    if summary:
        print(model.summary())
    # Fit network.
    return model, name_model



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm
from openpy_fxts.base_tf import tkm, tkl, tf


class encoder_lstm(tkm.Model):
    def __init__(self, n_future=None):
        super(encoder_lstm, self).__init__()
        self.lstm = tkl.LSTM(256, return_state=True)
        self.rep_vec = tkl.RepeatVector(n_future)

    def call(self, inputs, training=True, **kwargs):
        encoder_outputs1 = self.lstm(inputs)
        encoder_states1 = encoder_outputs1[1:]
        decoder_inputs = self.rep_vec(encoder_outputs1[0])
        return [decoder_inputs, encoder_states1]


class encoder_lstm2(tkm.Model):
    def __init__(self, n_future=None):
        super(encoder_lstm2, self).__init__()
        self.encoder_l1 = tkl.LSTM(128, return_sequences=True, return_state=True)
        self.encoder_l2 = tkl.LSTM(128, return_state=True)
        self.rep_vec = tkl.RepeatVector(n_future)

    def call(self, inputs, training=True, **kwargs):
        encoder_outputs1 = self.encoder_l1(inputs)
        encoder_states1 = encoder_outputs1[1:]
        encoder_outputs2 = self.encoder_l2(encoder_outputs1[0])
        encoder_states2 = encoder_outputs2[1:]
        decoder_inputs = self.rep_vec(encoder_outputs2[0])
        return [decoder_inputs, encoder_states1, encoder_states2]


class encoder_batch2_lstm2_drop2(tkm.Model):

    def __init__(self, n_future=None):
        super(encoder_batch2_lstm2_drop2, self).__init__()
        self.batch = tkl.BatchNormalization()
        self.lstm1 = tkl.LSTM(256, return_sequences=True)
        self.drop = tkl.Dropout(0.2)
        self.lstm2 = tkl.LSTM(256, return_sequences=False)
        self.rep_vec = tkl.RepeatVector(n_future)

    def call(self, inputs, training=True, **kwargs):
        encoder_input = self.batch(inputs)
        encoder_h1 = self.lstm1(encoder_input)
        encoder_h1 = self.drop(encoder_h1)
        encoder_out = self.lstm2(encoder_h1)
        encoder_out = self.drop(encoder_out)

        encoder_out = self.rep_vec(encoder_out)
        return encoder_out


class encoder_bilstm_state(tkm.Model):
    def __init__(self, n_future=None):
        super(encoder_bilstm_state, self).__init__()
        self.bilstm = tkl.Bidirectional(tkl.LSTM(256, return_state=True))
        self.rep_vec = tkl.RepeatVector(n_future)

    def call(self, inputs, training=True, **kwargs):
        encoder_outputs1 = self.bilstm(inputs)
        encoder_states1 = encoder_outputs1[1:]
        decoder_inputs = self.rep_vec(encoder_outputs1[0])
        return [decoder_inputs, encoder_states1]


class _encoder_bilstm_state_sequence(tkm.Model):
    def __init__(self, n_future=None):
        super(_encoder_bilstm_state_sequence, self).__init__()
        self.bilstm = tkl.Bidirectional(tkl.LSTM(128, dropout=0.2, return_state=True, return_sequences=True))
        self.concat = tkl.concatenate()
        self.rep_vec = tkl.RepeatVector(n_future)

    def call(self, inputs, training=True, **kwargs):
        encoder_stack_h, encoder_last_h_fw, encoder_last_c_fw, encoder_last_h_bw, encoder_last_c_bw = self.bilstm(inputs)
        encoder_last_h = tf.concat([encoder_last_h_fw, encoder_last_h_bw])
        encoder_last_c = tf.concat([encoder_last_c_fw, encoder_last_c_bw])
        decoder_inputs = self.rep_vec(encoder_last_h)
        return [decoder_inputs, encoder_last_h_fw, encoder_last_c_fw, encoder_last_h_bw, encoder_last_c_bw]


class encoder_bilstm2(tkm.Model):
    def __init__(self, n_future=None):
        super(encoder_bilstm2, self).__init__()
        self.encoder_l1 = tkl.Bidirectional(tkl.LSTM(128, return_sequences=True, return_state=True))
        self.encoder_l2 = tkl.Bidirectional(tkl.LSTM(128, return_state=True))
        self.rep_vec = tkl.RepeatVector(n_future)

    def call(self, inputs, training=True, **kwargs):
        encoder_outputs1 = self.encoder_l1(inputs)
        encoder_states1 = encoder_outputs1[1:]
        encoder_outputs2 = self.encoder_l2(encoder_outputs1[0])
        encoder_states2 = encoder_outputs2[1:]
        decoder_inputs = self.rep_vec(encoder_outputs2[0])
        return [decoder_inputs, encoder_states1, encoder_states2]
    

class encoder_conv1d(tkm.Model):
    def __init__(self, n_future=None):
        super(encoder_conv1d, self).__init__()
        self.conv1d1 = tkl.Conv1D(filters=64, kernel_size=11, activation='relu', padding='causal', dilation_rate=2)
        self.conv1d2 = tkl.Conv1D(filters=64, kernel_size=9, activation='relu', padding='causal')
        self.conv1d3 = tkl.Conv1D(filters=64, kernel_size=3, activation='relu')
        self.max_pool1d = tkl.MaxPooling1D(pool_size=2)
        self.flatten = tkl.Flatten()
        self.rep_vec = tkl.RepeatVector(n_future)

    def call(self, inputs, training=True, **kwargs):
        x = self.conv1d1(inputs)
        x = self.conv1d2(x)
        x = self.conv1d3(x)
        x = self.max_pool1d(x)
        x = self.flatten(x)
        x = self.rep_vec(x)
        return x


class encoder_multihead_conv1d_lstm(tkm.Model):
    def __init__(self, n_future=None, n_inp_ft=None, i=None):
        super(encoder_multihead_conv1d_lstm, self).__init__()
        self.n_inp_ft = n_inp_ft
        self.i = i
        self.conv1d = tkl.Conv1D(filters=64, kernel_size=3, activation='relu')
        self.max_pool1d = tkl.MaxPooling1D(pool_size=2)
        self.flatten = tkl.Flatten()
        self.reshape = tkl.Reshape((self.i[0].shape[1], n_inp_ft))
        self.concat = tkl.Concatenate(axis=1)
        self.lstm = tkl.LSTM(256, activation='relu')
        self.rep_vec = tkl.RepeatVector(n_future)


    def call(self, inputs, training=True, **kwargs):
        head_list = []
        for i in range(0, self.n_inp_ft):
            x = self.conv1d(inputs)
            x = self.flatten(x)
            head_list.append(x)

        x = self.concat(head_list)
        self.i = x
        x = self.reshape(x)
        x = self.lstm(x)
        x = self.rep_vec(x)
        return x



# -*- coding: utf-8 -*-
# @Time    : 14/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

class _get_dict_mdits:

    def __init__(self, name_mdl):
        self.name_mdl = name_mdl

    def config_mlp(self):
        if self.name_mdl == 'MLP_Dense':
            return {
                'input_layer': {
                    'units': 128,
                    'activation': None
                },
                'dense_layers': {
                    'units': [128, 64, 32],
                    'activations': ['relu', 'relu', 'relu']
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'output_layer': {
                    'activation': None
                }
            }

    def config_conv1d(self):
        # Simple models_fxts
        if self.name_mdl == 'Conv1D_Dense':
            return {
                'conv1d_layers': {
                    'filters': [64, 32],
                    'kernels': [9, 3],
                    'activations': ['relu', 'relu'],
                    'paddings': ['causal', 'causal'],
                    'pool_size': 2,
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'dense_layer': {
                    'activate': False,
                    'units': 32,
                    'activation': 'relu'
                },
                'output_layer': {
                    'activation': None
                }
            }
        if self.name_mdl == 'Multi_Conv1D_Dense':
            return {
                'conv1d_layers': {
                    'filters': [64, 32],
                    'kernels': [9, 3],
                    'activations': ['relu', 'relu', 'relu'],
                    'paddings': ['causal', 'causal', 'causal'],
                    'pool_size': 2,
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'dense_layer': {
                    'activate': False,
                    'units': 32,
                    'activation': 'relu'
                },
                'output_layer': {
                    'activation': None
                }
            }
        # Hybrids models_fxts
        if self.name_mdl == 'Conv1D_BiRNN':
            return {
                'conv1d_layers': {
                    'filters': [64, 32],
                    'kernels': [9, 3],
                    'activations': ['relu', 'relu', 'relu'],
                    'paddings': ['causal', 'causal', 'causal'],
                    'pool_size': 2
                },
                'rnn_layers': {
                    'type': 'GRU',
                    'units': [64, 32],
                    'activations': ['relu', 'relu'],
                    'sequences': [True, False],
                    'bidirectional': True
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'pool_size': 2,
                'dense_layer': {
                    'activate': False,
                    'units': 32,
                    'activation': 'relu'
                },
                'output_layer': {
                    'activation': None
                }
            }
        if self.name_mdl == 'TimeDist_Conv1D_BiRNN':
            return {
                'conv1d_layers': {
                    'filters': [64, 32],
                    'kernels': [9, 3],
                    'activations': ['relu', 'relu'],
                    'paddings': ['causal', 'causal'],
                    'maxpooling': True,
                    'pool_size': 2

                },
                'rnn_layer': {
                    'type': 'GRU',
                    'units': 256,
                    'activation': 'relu',
                    'sequences': False,
                    'bidirectional': False
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'pool_size': 2,
                'dense_layer': {
                    'units': 128,
                    'activation': 'relu'
                },
                'output_layer': {
                    'activation': None
                }
            }
        if self.name_mdl == 'Conv1D_BiRNN_Attention':
            return {
                'conv1d_layers': {
                    'filters': [64, 32],
                    'kernels': [9, 3],
                    'activations': ['relu', 'relu'],
                    'paddings': ['causal', 'causal'],
                    'pool_size': 2
                },
                'rnn_layers': {
                    'type': 'GRU',
                    'units': [64, 32],
                    'activations': ['relu', 'relu'],
                    'sequences': [True, True],
                    'bidirectional': True

                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'dense_layer': {
                    'activate': False,
                    'units': 32,
                    'activation': 'relu'
                },
                'output_layer': {
                    'activation': None
                }
            }

    def config_birnn(self):
        # Simple models_fxts
        if self.name_mdl == 'BiRNN_Dense':
            return {
                'rnn_layers': {
                    'type': 'GRU',
                    'units': [64, 32],
                    'activations': ['relu', 'relu'],
                    'sequences': [True, False],
                    'bidirectional': False
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'dense_layer': {
                    'activate': True,
                    'units': 128,
                    'activation': 'relu'
                },
                'output_layer': {
                    'activation': None
                }
            }
        if self.name_mdl == 'Multi_BiRNN_Dense':
            return {
                'multi_rnn_layers': {
                    'type': 'GRU',
                    'units': [64, 32],
                    'activations': ['relu', 'relu'],
                    'sequences': [True, False],
                    'bidirectional': True
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'rnn_layer': {
                    'units': 64,
                    'activation': 'relu',
                    'sequences': False
                },
                'dense_layer': {
                    'activate': True,
                    'units': 64,
                    'activation': 'relu'
                },
                'output_layer': {
                    'activation': None
                }
            }
        # Hybrid models_fxts
        if self.name_mdl == 'BiRNN_Conv1D':
            return {
                'rnn_layers': {
                    'type': 'GRU',
                    'units': [64, 32],
                    'activations': ['relu', 'relu'],
                    'sequences': [True, True],
                    'bidirectional': True
                },
                'conv1d_layers': {
                    'filters': [64, 32],
                    'kernels': [9, 2],
                    'activations': ['relu', 'relu'],
                    'paddings': ['causal', 'causal'],
                    'pool_size': 1
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'output_layer': {
                    'activation': None
                }
            }
        if self.name_mdl == 'BiRNN_TimeDist_Dense':
            return {
                'rnn_layer': {
                    'type': 'GRU',
                    'units': 64,
                    'activations': 'relu',
                    'sequences': True,
                    'bidirectional': True
                },
                'timedist_dense_layers': {
                    'units': [64, 32],
                    'activations': ['relu', 'relu']
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'dense_layer': {
                    'units': 64,
                    'activation': 'relu'
                },
                'output_layer': {
                    'activation': None
                }
            }
        if self.name_mdl == 'BiRNN_Bahdanau_Attention_Conv1D':
            return {
                'rnn_layers': {
                    'type': 'GRU',
                    'units': [64],
                    'activations': ['relu'],
                    'sequences': [True],
                    'bidirectional': False
                },
                'attention': {
                    'kernel_reg': 1e-4,
                    'bias_reg': 1e-4,
                    'reg_weight': 1e-4,
                },
                'conv1d_layers': {
                    'filters': [64],
                    'kernels': [9],
                    'activations': ['relu'],
                    'paddings': ['causal'],
                    'pool_size': 1
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'output_layer': {
                    'activation': None
                }
            }
        if self.name_mdl == 'BiRNN_Luong_Attention_Conv1D':
            return {
                'rnn_layers': {
                    'type': 'GRU',
                    'units': [64],
                    'activations': ['relu'],
                    'sequences': [True],
                    'bidirectional': False
                },
                'attention': {
                    'kernel_reg': 1e-4,
                    'bias_reg': 1e-4,
                    'reg_weight': 1e-4,
                },
                'conv1d_layers': {
                    'filters': [64],
                    'kernels': [9],
                    'activations': ['relu'],
                    'paddings': ['causal'],
                    'pool_size': 1
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'output_layer': {
                    'activation': None
                }
            }
        if self.name_mdl == 'BiRNN_MultiHeadAttention_Conv1D':
            return {
                'rnn_layers': {
                    'type': 'GRU',
                    'units': [64],
                    'activations': ['relu'],
                    'sequences': [True],
                    'bidirectional': False
                },
                'attention': {
                    'num_heads': 2,
                    'key_dim': 2,
                    'dropout': 0.1
                },
                'conv1d_layers': {
                    'filters': [64],
                    'kernels': [9],
                    'activations': ['relu'],
                    'paddings': ['causal'],
                    'pool_size': 1
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
                'output_layer': {
                    'activation': None
                }
            }

    def config_encdec(self):
        if self.name_mdl == 'EncDec_BiRNN':
            return {
                'encoder': {
                    'rnn_layers': {
                        'type': 'GRU',
                        'units': [64, 32],
                        'activations': ['relu', 'relu'],
                        'sequences': [True, False],
                        'bidirectional': True
                    }
                },
                'decoder': {
                    'rnn_layers': {
                        'type': 'GRU',
                        'units': [64, 32],
                        'activations': ['relu', 'relu'],
                        'sequences': [True, True],
                        'bidirectional': True
                    }
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
            }
        if self.name_mdl == 'EncDec_Conv1D_BiRNN':
            return {
                'encoder': {
                    'conv1d_layers': {
                        'filters': [64, 32],
                        'kernels': [9, 3],
                        'activations': ['relu', 'relu'],
                        'paddings': ['causal', 'causal'],
                        'pool_size': 2
                    },
                },
                'decoder': {
                    'rnn_layers': {
                        'type': 'GRU',
                        'units': [64, 32],
                        'activations': ['relu', 'relu'],
                        'sequences': [True, True],
                        'bidirectional': True
                    }
                },
                'dropout': {
                    'activate': True,
                    'rate': 0.3
                },
            }


from openpy_fxts.mdits.mdls_dl.base_tf import K, tkl, tkm, tf

import keras.layers.core
from typing import Tuple

feat_1 = 1  # index+1 where feature is ending
feat_2 = 2  # index+1 where feature is ending
feat_3 = 3  # index+1 where feature is ending
feat_4 = 4  # index+1 where feature is ending
# adding 0 at the front to help in the very first iteration while creating layers in a loop
feat_lst = [0, feat_1, feat_2, feat_3, feat_4]


# Define an input sequence and process it.
class _Luongs_concat_attention(tkl.Layer):

    def __init__(self, units):
        super().__init__()
        self.units = units

    def build(self, input_shape):
        self.w1 = self.add_weight(
            "w1",  # creating trainable matrix of size ( hidden size * downsample  )
            shape=(self.units, int(self.units / 2)),
            initializer="random_normal",
            trainable=True, dtype='float64')

        self.w2 = self.add_weight(
            "w2",  # creating trainable matrix of size ( hidden size * downsample  )
            shape=(self.units, int(self.units / 2)),
            initializer="random_normal",
            trainable=True, dtype='float64')

        self.V = self.add_weight(
            "V",  # creating trainable matrix of size ( 1 * hidden size  )
            shape=(1, self.units),
            initializer="random_normal",
            trainable=True, dtype='float64')

    def call(self, query, values):
        query = tf.cast(query, 'float64')
        query_with_time_axis = tf.expand_dims(query, 1)
        enc_w1 = tf.linalg.matmul(values, self.w1, transpose_a=False, transpose_b=False)
        dec_w2 = tf.linalg.matmul(query_with_time_axis, self.w2, transpose_a=False, transpose_b=False)
        dec_w2 = tf.broadcast_to(dec_w2, tf.shape(enc_w1))
        score = tf.math.tanh(tf.concat([enc_w1, dec_w2], axis=-1))
        score = tf.linalg.matmul(score, self.V, transpose_a=False, transpose_b=True)
        attention_weights = tf.nn.softmax(score, axis=1)
        context_vector = attention_weights * values
        context_vector = tf.reduce_sum(context_vector, axis=1)

        return context_vector, attention_weights


# without teacher forcing
# Define an input sequence and process it.
class _Luongs_dot_attention(tkl.Layer):

    def __init__(self):
        super().__init__()

    def call(self, query, values):
        query = tf.cast(query, 'float64')
        query_with_time_axis = tf.expand_dims(query, 1)
        score = tf.linalg.matmul(values, query_with_time_axis, transpose_a=False, transpose_b=True)
        attention_weights = tf.nn.softmax(score, axis=1)
        context_vector = attention_weights * values
        context_vector = tf.reduce_sum(context_vector, axis=1)
        return context_vector, attention_weights


# without teacher forcing
# Define an input sequence and process it.
class _Luongs_general_attention(tkl.Layer):

    def __init__(self, units):
        super().__init__()
        self.units = units

    def build(self, input_shape):
        self.w = self.add_weight(  # creating trainable matrix of size ( hidden size * hidden size  )
            shape=(self.units, self.units),
            initializer="random_normal",
            trainable=True, dtype='float64')

    def call(self, query, values):
        query = tf.cast(query, 'float64')
        query_with_time_axis = tf.expand_dims(query, 1)
        score = tf.linalg.matmul(values, self.w, transpose_a=False, transpose_b=False)
        score = tf.linalg.matmul(score, query_with_time_axis, transpose_a=False, transpose_b=True)
        attention_weights = tf.nn.softmax(score, axis=1)
        context_vector = attention_weights * values
        context_vector = tf.reduce_sum(context_vector, axis=1)
        return context_vector, attention_weights


# Experiment Nro. 1
# Without teacher forcing
# Define an input sequence and process it.
class MyModel_exp1_hybrid_normal(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):

        super().__init__()
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features
        var = []
        for i in range(features):
            lst = []
            encoder = tkl.LSTM(units, return_state=True, dtype='float64')
            decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
            dense = tkl.Dense(1, dtype='float64')
            lst.append(encoder)
            lst.append(decoder)
            lst.append(dense)
            var.append(lst)
        self.var = var

    def call(self, data):
        encoder_input = data
        all_outputs = []
        for i in range(self.features):
            outputs = []
            enc_input = tf.expand_dims(encoder_input[:, :, i], 2)
            e_input = enc_input
            _, hidden, cell = self.var[i][0](e_input)  # We discard `encoder_outputs` and only keep the states.
            states = [hidden, cell]
            cur_vec = tf.expand_dims(enc_input[:, -1, :], 1)
            final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
            final = self.var[i][2](final)
            outputs.append(final)
            states = [hidden, cell]
            for j in range(self.decoder_outputs_length - 1):
                cur_vec = final
                final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
                final = self.var[i][2](final)
                outputs.append(final)
                states = [hidden, cell]
            outputs = tf.concat(outputs, axis=1)
            all_outputs.append(outputs)
        all_outputs = tf.concat(all_outputs, axis=2)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp1_hybrid_tf(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features
        var = []
        for i in range(len(feat_lst) - 1):
            lst = []
            encoder = tkl.LSTM(units, return_state=True, dtype='float64')
            decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
            dense = tkl.Dense(feat_lst[i + 1] - feat_lst[i], dtype='float64')
            lst.append(encoder)
            lst.append(decoder)
            lst.append(dense)
            var.append(lst)
        self.var = var

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        all_outputs = []
        for i in range(len(feat_lst) - 1):
            outputs = []
            enc_input, d_input = encoder_input[:, :, feat_lst[i]:feat_lst[i + 1]], decoder_input[:, :,
                                                                                   feat_lst[i]:feat_lst[i + 1]]
            e_input = enc_input
            _, hidden, cell = self.var[i][0](e_input)  # We discard `encoder_outputs` and only keep the states.
            states = [hidden, cell]
            cur_vec = tf.expand_dims(enc_input[:, -1, :], 1)
            final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
            final = self.var[i][2](final)
            outputs.append(final)
            states = [hidden, cell]
            for j in range(self.decoder_outputs_length - 1):
                cur_vec = tf.expand_dims(d_input[:, j, :], 1)
                final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
                final = self.var[i][2](final)
                outputs.append(final)
                states = [hidden, cell]
            outputs = tf.concat(outputs, axis=1)
            all_outputs.append(outputs)
        all_outputs = tf.concat(all_outputs, axis=2)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp1_separate_tf(tkm.Model):
    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):

        super().__init__()
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features
        var = []
        for i in range(features):
            lst = []
            encoder = tkl.LSTM(units, return_state=True, dtype='float64')
            decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
            dense = tkl.Dense(1, dtype='float64')
            lst.append(encoder)
            lst.append(decoder)
            lst.append(dense)
            var.append(lst)
        self.var = var

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        all_outputs = []
        for i in range(self.features):
            outputs = []
            enc_input, d_input = tf.expand_dims(encoder_input[:, :, i], 2), tf.expand_dims(decoder_input[:, :, i], 2)
            e_input = enc_input
            _, hidden, cell = self.var[i][0](e_input)  # We discard `encoder_outputs` and only keep the states.
            states = [hidden, cell]
            cur_vec = tf.expand_dims(enc_input[:, -1, :], 1)
            final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
            final = self.var[i][2](final)
            outputs.append(final)
            states = [hidden, cell]
            for j in range(self.decoder_outputs_length - 1):
                cur_vec = tf.expand_dims(d_input[:, j, :], 1)
                final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
                final = self.var[i][2](final)
                outputs.append(final)
                states = [hidden, cell]
            outputs = tf.concat(outputs, axis=1)
            all_outputs.append(outputs)
        all_outputs = tf.concat(all_outputs, axis=2)
        return all_outputs


# without teacher forcing
# Define an input sequence and process it.
class MyModel_exp1_separate_normal(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):

        super().__init__()
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features
        var = []
        for i in range(features):
            lst = []
            encoder = tkl.LSTM(units, return_state=True, dtype='float64')
            decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
            dense = tkl.Dense(1, dtype='float64')
            lst.append(encoder)
            lst.append(decoder)
            lst.append(dense)
            var.append(lst)

        self.var = var

    def call(self, data):
        encoder_input = data
        all_outputs = []
        for i in range(self.features):
            outputs = []
            enc_input = tf.expand_dims(encoder_input[:, :, i], 2)
            e_input = enc_input
            _, hidden, cell = self.var[i][0](e_input)  # We discard `encoder_outputs` and only keep the states.
            states = [hidden, cell]
            cur_vec = tf.expand_dims(enc_input[:, -1, :], 1)
            final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
            final = self.var[i][2](final)
            outputs.append(final)
            states = [hidden, cell]
            for j in range(self.decoder_outputs_length - 1):
                cur_vec = final
                final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
                final = self.var[i][2](final)
                outputs.append(final)
                states = [hidden, cell]
            outputs = tf.concat(outputs, axis=1)
            all_outputs.append(outputs)
        all_outputs = tf.concat(all_outputs, axis=2)
        return all_outputs


# without teacher forcing
# Define an input sequence and process it.
class MyModel_exp1_shared_normal(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_state=True, dtype='float64')
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        all_outputs = []
        _, hidden, cell = self.encoder(encoder_input)  # We discard `encoder_outputs` and only keep the states.
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            cur_vec = final
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


# without teacher forcing
class MyModel_exp1_shared_normal_luongs_dot_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_dot_attention()
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(encoder_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = final
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


class MyModel_exp1_shared_normal_luongs_general_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_general_attention(units=units)
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(encoder_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = final
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


class MyModel_exp1_shared_normal_luongs_concat_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_concat_attention(units=units)
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(encoder_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = final
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


class MyModel_exp1_shared_tf(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_state=True, dtype='float64')
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        all_outputs = []
        _, hidden, cell = self.encoder(encoder_input)  # We discard `encoder_outputs` and only keep the states.
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            cur_vec = tf.expand_dims(decoder_input[:, i, :], 1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


class MyModel_exp1_shared_tf_luongs_general_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_general_attention(units=units)
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(encoder_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = tf.expand_dims(decoder_input[:, i, :], 1)
            cur_vec = tf.cast(cur_vec, 'float64')
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp1_shared_tf_luongs_concat_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_concat_attention(units=units)
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(encoder_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = tf.expand_dims(decoder_input[:, i, :], 1)
            cur_vec = tf.cast(cur_vec, 'float64')
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp1_shared_tf_luongs_dot_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_dot_attention()
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(encoder_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = tf.expand_dims(decoder_input[:, i, :], 1)
            cur_vec = tf.cast(cur_vec, 'float64')
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


# Experiment Nro. 2
class MyModel_exp2_hybrid_normal(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features
        var = []
        for i in range(len(feat_lst) - 1):
            lst = []
            encoder = tkl.LSTM(units, return_state=True, dtype='float64')
            decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
            dense = tkl.Dense(feat_lst[i + 1] - feat_lst[i], dtype='float64')
            lst.append(encoder)
            lst.append(decoder)
            lst.append(dense)
            var.append(lst)
        self.var = var

    def call(self, data):
        encoder_input = data
        all_outputs = []
        for i in range(len(feat_lst) - 1):
            outputs = []
            enc_input = encoder_input[:, :, feat_lst[i]:feat_lst[i + 1]]
            e_input = enc_input[:, :-1, :]
            _, hidden, cell = self.var[i][0](e_input)  # We discard `encoder_outputs` and only keep the states.
            states = [hidden, cell]
            cur_vec = tf.expand_dims(enc_input[:, -1, :], 1)
            final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
            final = self.var[i][2](final)
            outputs.append(final)
            states = [hidden, cell]
            for j in range(self.decoder_outputs_length - 1):
                cur_vec = final
                final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
                final = self.var[i][2](final)
                outputs.append(final)
                states = [hidden, cell]
            outputs = tf.concat(outputs, axis=1)
            all_outputs.append(outputs)
        all_outputs = tf.concat(all_outputs, axis=2)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp2_hybrid_tf(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features
        var = []
        for i in range(len(feat_lst) - 1):
            lst = []
            encoder = tkl.LSTM(units, return_state=True, dtype='float64')
            decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
            dense = tkl.Dense(feat_lst[i + 1] - feat_lst[i], dtype='float64')
            lst.append(encoder)
            lst.append(decoder)
            lst.append(dense)
            var.append(lst)
        self.var = var

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        all_outputs = []
        for i in range(len(feat_lst) - 1):
            outputs = []
            enc_input, d_input = encoder_input[:, :, feat_lst[i]:feat_lst[i + 1]], decoder_input[:, :,
                                                                                   feat_lst[i]:feat_lst[i + 1]]
            e_input = enc_input[:, :-1, :]
            _, hidden, cell = self.var[i][0](e_input)  # We discard `encoder_outputs` and only keep the states.
            states = [hidden, cell]
            cur_vec = tf.expand_dims(enc_input[:, -1, :], 1)
            final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
            final = self.var[i][2](final)
            outputs.append(final)
            states = [hidden, cell]
            for j in range(self.decoder_outputs_length - 1):
                cur_vec = tf.expand_dims(d_input[:, j, :], 1)
                final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
                final = self.var[i][2](final)
                outputs.append(final)
                states = [hidden, cell]
            outputs = tf.concat(outputs, axis=1)
            all_outputs.append(outputs)
        all_outputs = tf.concat(all_outputs, axis=2)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp2_separate_tf(tkm.Model):
    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features
        var = []
        for i in range(features):
            lst = []
            encoder = tkl.LSTM(units, return_state=True, dtype='float64')
            decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
            dense = tkl.Dense(1, dtype='float64')
            lst.append(encoder)
            lst.append(decoder)
            lst.append(dense)
            var.append(lst)
        self.var = var

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        all_outputs = []
        for i in range(self.features):
            outputs = []
            enc_input, d_input = tf.expand_dims(encoder_input[:, :, i], 2), tf.expand_dims(decoder_input[:, :, i], 2)
            e_input = enc_input[:, :-1, :]
            _, hidden, cell = self.var[i][0](e_input)  # We discard `encoder_outputs` and only keep the states.
            states = [hidden, cell]
            cur_vec = tf.expand_dims(enc_input[:, -1, :], 1)
            final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
            final = self.var[i][2](final)
            outputs.append(final)
            states = [hidden, cell]
            for j in range(self.decoder_outputs_length - 1):
                cur_vec = tf.expand_dims(d_input[:, j, :], 1)
                final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
                final = self.var[i][2](final)
                outputs.append(final)
                states = [hidden, cell]
            outputs = tf.concat(outputs, axis=1)
            all_outputs.append(outputs)
        all_outputs = tf.concat(all_outputs, axis=2)
        return all_outputs


# without teacher forcing
# Define an input sequence and process it.
class MyModel_exp2_separate_normal(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features
        var = []
        for i in range(features):
            lst = []
            encoder = tkl.LSTM(units, return_state=True, dtype='float64')
            decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
            dense = tkl.Dense(1, dtype='float64')
            lst.append(encoder)
            lst.append(decoder)
            lst.append(dense)
            var.append(lst)
        self.var = var

    def call(self, data):
        encoder_input = data
        all_outputs = []
        for i in range(self.features):
            outputs = []
            enc_input = tf.expand_dims(encoder_input[:, :, i], 2)
            e_input = enc_input[:, :-1, :]
            _, hidden, cell = self.var[i][0](e_input)  # We discard `encoder_outputs` and only keep the states.
            states = [hidden, cell]
            cur_vec = tf.expand_dims(enc_input[:, -1, :], 1)
            final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
            final = self.var[i][2](final)
            outputs.append(final)
            states = [hidden, cell]
            for j in range(self.decoder_outputs_length - 1):
                cur_vec = final
                final, hidden, cell = self.var[i][1](cur_vec, initial_state=states)
                final = self.var[i][2](final)
                outputs.append(final)
                states = [hidden, cell]
            outputs = tf.concat(outputs, axis=1)
            all_outputs.append(outputs)
        all_outputs = tf.concat(all_outputs, axis=2)
        return all_outputs


# without teacher forcing
# Define an input sequence and process it.
class MyModel_exp2_shared_normal(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_state=True, dtype='float64')
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        enc_input = encoder_input[:, :-1, :]
        all_outputs = []
        _, hidden, cell = self.encoder(enc_input)  # We discard `encoder_outputs` and only keep the states.
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            cur_vec = final
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs

    # without teacher forcing


class MyModel_exp2_shared_normal_luongs_dot_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_dot_attention()
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        enc_input = encoder_input[:, :-1, :]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(enc_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = final
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


class MyModel_exp2_shared_normal_luongs_general_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_general_attention(units=units)
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        enc_input = encoder_input[:, :-1, :]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(enc_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = final
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)

        return all_outputs


class MyModel_exp2_shared_normal_luongs_concat_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_concat_attention(units=units)
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        enc_input = encoder_input[:, :-1, :]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(enc_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = final
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


class MyModel_exp2_shared_tf(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_state=True, dtype='float64')
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        enc_input = encoder_input[:, :-1, :]
        all_outputs = []
        _, hidden, cell = self.encoder(enc_input)  # We discard `encoder_outputs` and only keep the states.
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            cur_vec = tf.expand_dims(decoder_input[:, i, :], 1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


class MyModel_exp2_shared_tf_luongs_general_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()

        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_general_attention(units=units)
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        enc_input = encoder_input[:, :-1, :]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(enc_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = tf.expand_dims(decoder_input[:, i, :], 1)
            cur_vec = tf.cast(cur_vec, 'float64')
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp2_shared_tf_luongs_concat_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_concat_attention(units=units)
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        enc_input = encoder_input[:, :-1, :]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(enc_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = tf.expand_dims(decoder_input[:, i, :], 1)
            cur_vec = tf.cast(cur_vec, 'float64')
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp2_shared_tf_luongs_dot_attention(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.attention = _Luongs_dot_attention()
        self.decoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        enc_input = encoder_input[:, :-1, :]
        all_outputs = []
        encoder_output, hidden, cell = self.encoder(enc_input)
        context_vector, attention_weights = self.attention(hidden, encoder_output)
        states = [hidden, cell]
        cur_vec = tf.expand_dims(encoder_input[:, -1, :], 1)
        cur_vec = tf.cast(cur_vec, 'float64')
        context_vector = tf.expand_dims(context_vector, 1)
        cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
        final, hidden, cell = self.decoder(cur_vec, initial_state=states)
        final = self.dense(final)
        all_outputs.append(final)
        states = [hidden, cell]
        for i in range(self.decoder_outputs_length - 1):
            context_vector, attention_weights = self.attention(hidden, encoder_output)
            cur_vec = tf.expand_dims(decoder_input[:, i, :], 1)
            cur_vec = tf.cast(cur_vec, 'float64')
            context_vector = tf.expand_dims(context_vector, 1)
            cur_vec = tf.concat([context_vector, cur_vec], axis=-1)
            final, hidden, cell = self.decoder(cur_vec, initial_state=states)
            final = self.dense(final)
            all_outputs.append(final)
            states = [hidden, cell]
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


# Define an input sequence and process it.
class MyModel_exp4_normal(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.decoder = tkl.LSTM(units, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input = data
        encoder_input = tf.cast(encoder_input, dtype='float64')
        encoder_outputs, hidden, cell = self.encoder(encoder_input)
        states = [hidden, cell]
        all_outputs = []
        final = self.decoder(encoder_outputs, initial_state=states)
        final = self.dense(final)
        final = tf.expand_dims(final, axis=1)
        all_outputs.append(final)
        for i in range(self.decoder_outputs_length - 1):
            encoder_input = encoder_input[:, 1:, :]
            encoder_input = tf.concat([encoder_input, all_outputs[-1]], axis=1)
            encoder_outputs, hidden, cell = self.encoder(encoder_input)
            states = [hidden, cell]
            final = self.decoder(encoder_outputs, initial_state=states)
            final = self.dense(final)
            final = tf.expand_dims(final, axis=1)
            all_outputs.append(final)
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


# teacher forcing
# Define an input sequence and process it.
class MyModel_exp4_tf(tkm.Model):

    def __init__(self, encoder_inputs_length, decoder_outputs_length, features, units):
        super().__init__()
        self.encoder = tkl.LSTM(units, return_sequences=True, return_state=True, dtype='float64')
        self.decoder = tkl.LSTM(units, dtype='float64')
        self.dense = tkl.Dense(features, dtype='float64')
        self.decoder_outputs_length = decoder_outputs_length
        self.features = features

    def call(self, data):
        encoder_input, decoder_input = data[0], data[1]
        encoder_outputs, hidden, cell = self.encoder(encoder_input)
        states = [hidden, cell]
        all_outputs = []
        final = self.decoder(encoder_outputs, initial_state=states)
        final = self.dense(final)
        final = tf.expand_dims(final, axis=1)
        all_outputs.append(final)
        for i in range(self.decoder_outputs_length - 1):
            encoder_input = encoder_input[:, 1:, :]
            encoder_input = tf.concat([encoder_input, decoder_input[:, i:i + 1, :]], axis=1)
            encoder_outputs, hidden, cell = self.encoder(encoder_input)
            states = [hidden, cell]
            final = self.decoder(encoder_outputs, initial_state=states)
            final = self.dense(final)
            final = tf.expand_dims(final, axis=1)
            all_outputs.append(final)
        all_outputs = tf.concat(all_outputs, axis=1)
        return all_outputs


class InnerConv1DBlock(tkl.Layer):

    def __init__(self, filters: int, h: float, kernel_size: int, neg_slope: float = .01, dropout: float = .5, **kwargs):
        if filters <= 0 or h <= 0:
            raise ValueError('filters and h must be positive')
        super(InnerConv1DBlock, self).__init__(**kwargs)
        self.conv1d = tkl.Conv1D(max(round(h * filters), 1), kernel_size, padding='same')
        self.leakyrelu = tkl.LeakyReLU(neg_slope)
        self.dropout = tkl.Dropout(dropout)
        self.conv1d2 = tkl.Conv1D(filters, kernel_size, padding='same')
        self.tanh = tf.keras.activations.tanh

    def call(self, input_tensor, training=None):
        x = self.conv1d(input_tensor)
        x = self.leakyrelu(x)

        if training:
            x = self.dropout(x)

        x = self.conv1d2(x)
        x = self.tanh(x)
        return x


class SciBlock(tkl.Layer):
    def __init__(self, features: int, kernel_size: int, h: int, **kwargs):
        """
        :param features: number of features in the output
        :param kernel_size: kernel size of the convolutional layers
        :param h: scaling factor for convolutional module
        """

        super(SciBlock, self).__init__(**kwargs)
        self.features = features
        self.kernel_size = kernel_size
        self.h = h

    def build(self, input_shape):
        self.conv1ds = {k: InnerConv1DBlock(filters=self.features, h=self.h, kernel_size=self.kernel_size, name=k)
                        for k in ['psi', 'phi', 'eta', 'rho']}  # regularize?
        super().build(input_shape)
        # [layer.build(input_shape) for layer in self.conv1ds.values()]  # unneeded?

    def call(self, inputs, training=None):
        F_odd, F_even = inputs[:, ::2], inputs[:, 1::2]

        # Interactive learning as described in the paper
        F_s_odd = F_odd * tf.math.exp(self.conv1ds['phi'](F_even))
        F_s_even = F_even * tf.math.exp(self.conv1ds['psi'](F_odd))

        F_prime_odd = F_s_odd + self.conv1ds['rho'](F_s_even)
        F_prime_even = F_s_even - self.conv1ds['eta'](F_s_odd)

        return F_prime_odd, F_prime_even


class Interleave(tkl.Layer):
    """A layer used to reverse the even-odd split operation."""

    def __init__(self, **kwargs):
        super(Interleave, self).__init__(**kwargs)

    def interleave(self, slices):
        if not slices:
            return slices
        elif len(slices) == 1:
            return slices[0]

        mid = len(slices) // 2
        even = self.interleave(slices[:mid])
        odd = self.interleave(slices[mid:])

        shape = tf.shape(even)
        return tf.reshape(tf.stack([even, odd], axis=3), (shape[0], shape[1] * 2, shape[2]))

    def call(self, inputs):
        return self.interleave(inputs)


class SciNet(tkl.Layer):
    def __init__(
            self,
            horizon: int,
            features: int,
            levels: int,
            h: int,
            kernel_size: int,
            regularizer: Tuple[float, float] = (0, 0),
            **kwargs
    ):
        """
        :param horizon: number of time stamps in output
        :param features: number of features in output
        :param levels: height of the binary tree + 1
        :param h: scaling factor for convolutional module in each SciBlock
        :param kernel_size: kernel size of convolutional module in each SciBlock
        :param regularizer: activity regularization (not implemented)
        """

        if levels < 1:
            raise ValueError('Must have at least 1 level')
        super(SciNet, self).__init__(**kwargs)
        self.horizon = horizon
        self.features = features
        self.levels = levels
        self.interleave = Interleave()
        self.flatten = tkl.Flatten()
        self.dense = tkl.Dense(
            horizon * features,
            kernel_regularizer=tf.keras.regularizers.L1L2(0.001, 0.01),
            # activity_regularizer=L1L2(0.001, 0.01)
        )
        # self.regularizer = tkl.ActivityRegularization(l1=regularizer[0], l2=regularizer[1])

        # tree of sciblocks
        self.sciblocks = [SciBlock(features=features, kernel_size=kernel_size, h=h)
                          for _ in range(2 ** levels - 1)]

    def build(self, input_shape):
        if input_shape[1] / 3 ** self.levels % 1 != 0:
            raise ValueError(f'timestamps {input_shape[1]} must be evenly divisible by a tree with '
                             f'{self.levels} levels')
        super().build(input_shape)
        # [layer.build(input_shape) for layer in self.sciblocks]  # input_shape

    def call(self, inputs, training=None):
        # cascade input down a binary tree of sci-blocks
        lvl_inputs = [inputs]  # inputs for current level of the tree
        for i in range(self.levels):
            i_end = 2 ** (i + 1) - 1
            i_start = i_end - 2 ** i
            lvl_outputs = [output for j, tensor in zip(range(i_start, i_end), lvl_inputs)
                           for output in self.sciblocks[j](tensor)]
            lvl_inputs = lvl_outputs

        x = self.interleave(lvl_outputs)
        x += inputs

        # not sure if this is the correct way of doing it. The paper merely said to use a fully connected layer to
        # produce an output. Can't use TimeDistributed wrapper. It would force the layer's timestamps to match that of
        # the input -- something SCINet is supposed to solve
        x = self.flatten(x)
        x = self.dense(x)
        x = tf.reshape(x, (-1, self.horizon, self.features))

        return x


class StackedSciNet(tkl.Layer):
    def __init__(
            self,
            horizon: int,
            features: int,
            stacks: int,
            levels: int,
            h: int,
            kernel_size: int,
            regularizer: Tuple[float, float] = (0, 0),
            **kwargs
    ):
        """
        :param horizon: number of time stamps in output
        :param stacks: number of stacked SciNets
        :param levels: number of levels for each SciNet
        :param h: scaling factor for convolutional module in each SciBlock
        :param kernel_size: kernel size of convolutional module in each SciBlock
        :param regularizer: activity regularization (not implemented)
        """

        if stacks < 1:
            raise ValueError('Must have at least 1 stack')
        super(StackedSciNet, self).__init__(**kwargs)
        self.horizon = horizon
        self.scinets = [
            SciNet(
                horizon=horizon,
                features=features,
                levels=levels,
                h=h,
                kernel_size=kernel_size,
                regularizer=regularizer
            ) for _ in range(stacks)
        ]
        # self.mse_fn = tf.keras.metrics.MeanSquaredError()
        # self.mae_fn = tf.keras.metrics.MeanAbsoluteError()

    # def build(self, input_shape):
    #     super().build(input_shape)
    #     [stack.build(input_shape) for stack in self.scinets]

    def call(self, inputs, targets=None, sample_weights=None, training=None):
        outputs = []
        for scinet in self.scinets:
            x = scinet(inputs)
            outputs.append(x)  # keep each stack's output for intermediate supervision
            inputs = tf.concat([x, inputs[:, x.shape[1]:, :]], axis=1)

        if targets is not None:
            # Calculate metrics
            # mse = self.mse_fn(targets, x, sample_weights)
            # mae = self.mae_fn(targets, x, sample_weights)
            # self.add_metric(mse, name='mean_squared_error')
            # self.add_metric(mae, name='mean_absolute_error')

            if training:
                # Calculate loss as sum of mean of norms of differences between output and input feature vectors for
                # each stack
                stacked_outputs = tf.stack(outputs)
                differences = stacked_outputs - targets
                loss = tf.linalg.normalize(differences, axis=1)[1]
                loss = tf.reshape(loss, (-1, self.horizon))
                loss = tf.reduce_sum(loss, 1)
                loss = loss / self.horizon
                loss = tf.reduce_sum(loss)
                self.add_loss(loss)

        return x

    def get_config(self):
        return super().get_config()


SINGLE_ATTENTION_VECTOR = False
def attention_3d_block(inputs):
    # inputs.shape = (batch_size, time_steps, input_dim)
    input_dim = int(inputs.shape[2])
    a = inputs
    # a = Permute((2, 1))(inputs)
    # a = Reshape((input_dim, TIME_STEPS))(a) # this line is not useful. It's just to know which dimension is what.
    a = tkl.Dense(input_dim, activation='softmax')(a)
    if SINGLE_ATTENTION_VECTOR:
        a = keras.layers.core.Lambda(lambda x: K.mean(x, axis=1), name='dim_reduction')(a)
        a = tkl.RepeatVector(input_dim)(a)
    a_probs = tkl.Permute((1, 2), name='attention_vec')(a)

    # output_attention_mul = merge([inputs, a_probs], name='attention_mul', mode='mul')
    output_attention_mul = tkl.Multiply()([inputs, a_probs])
    return output_attention_mul



# -*- coding: utf-8 -*-
# @Time    : 10/07/2023
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from openpy_fxts.base_tf import tkm, tkl, tkr, value_miss
from openpy_fxts.architectures.layers.attention_class import SeqSelfAttention


class bilstm_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(bilstm_dense, self).__init__()
        if config is None:
            config = {
                'bilstm': {
                    'units': 256,
                    'activation': 'tanh',
                    'recurrent_activation': 'relu'
                },
                'dropout': 0.3,
                'dense': {
                    'activation': 'linear'
                }
            }

        self.mask = tkl.Masking(mask_value=value_miss)
        self.bilstm = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm']['units'],
                activation=config['bilstm']['activation'],
                recurrent_activation=config['bilstm']['recurrent_activation'],
                return_sequences=False
            )
        )
        self.dropout = tkl.Dropout(config['dropout'])
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.bilstm(x)
        x = self.dropout(x)
        x = self.dense1(x)
        x = self.dense2(x)
        return x


def bilstm_dense_model(
        n_past=None,
        n_inp_ft=None,
        n_future=None,
        n_out_ft=None):

    input_layer = tkl.Input(shape=(n_past, n_inp_ft))
    mask = tkl.Masking(mask_value=-1)(input_layer)
    bilstm = tkl.Bidirectional(tkl.LSTM(units=256, return_sequences=True))(mask)
    conv1d_1 = tkl.Conv1D(filters=64, kernel_size=3, activation='relu', padding='same')(bilstm)
    glob_pool = tkl.GlobalAveragePooling1D()(conv1d_1)
    drop = tkl.Dropout(0.3)(glob_pool)
    dense = tkl.Dense(units=(n_future * n_out_ft), activation='linear')(drop)
    reshape = tkl.Reshape((n_future, n_out_ft))(dense)
    out_layer = tkl.Conv1D(n_out_ft, 1, padding='same')(reshape)

    return tkm.Model(inputs=input_layer, outputs=out_layer)


class bilstm_conv1d_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(bilstm_conv1d_dense, self).__init__()
        if config is None:
            config = {
                'bilstm': {
                    'units': 256,
                    },
                'conv1d': {
                    'filter': 64,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'same'
                },
                'dropout': 0.3,
                'dense': {
                    'activation': 'relu'
                }
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.bilstm = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm']['units'],
                return_sequences=True
            )
        )
        self.conv1 = tkl.Conv1D(
            filters=config['conv1d']['filter'],
            kernel_size=config['conv1d']['kernel'],
            activation=config['conv1d']['activation'],
            padding=config['conv1d']['padding']
        )
        self.pool1d = tkl.GlobalAveragePooling1D()
        self.dropout = tkl.Dropout(config['dropout'])
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))
        self.conv2 = tkl.Conv1D(n_out_ft, 1, padding='same')

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.bilstm(x)
        x = self.conv1(x)
        x = self.pool1d(x)
        x = self.dropout(x)
        x = self.dense1(x)
        x = self.dense2(x)
        x = self.conv2(x)
        return x


class bilstm_bahdanau_attention_conv1d_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(bilstm_bahdanau_attention_conv1d_dense, self).__init__()
        if config is None:
            config = {
                'bilstm1': {'units': 64},
                'conv1d_1': {
                    'filter': 128,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'same'
                },
                'dropout': 0.5,
                'bilstm2': {'units': 32},
                'conv1d_2': {
                    'filter': 256,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'same'
                },
                'dense': {'activation': 'relu'}
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.bilstm1 = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm1']['units'],
                return_sequences=True
            )
        )
        self.attention1 = SeqSelfAttention(
            attention_type=SeqSelfAttention.ATTENTION_TYPE_ADD,
            kernel_regularizer=tkr.l2(1e-4),
            bias_regularizer=tkr.l1(1e-4),
            attention_regularizer_weight=1e-4,
            name='Attention1'
        )
        self.conv1 = tkl.Conv1D(
            filters=config['conv1d_1']['filter'],
            kernel_size=config['conv1d_1']['kernel'],
            activation=config['conv1d_1']['activation'],
            padding=config['conv1d_1']['padding']
        )
        self.maxpool1d = tkl.MaxPool1D()
        self.bilstm2 = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm2']['units'],
                return_sequences=True
            )
        )
        self.attention2 = SeqSelfAttention(
            attention_type=SeqSelfAttention.ATTENTION_TYPE_ADD,
            kernel_regularizer=tkr.l2(1e-4),
            bias_regularizer=tkr.l1(1e-4),
            attention_regularizer_weight=1e-4,
            name='Attention2'
        )
        self.conv2 = tkl.Conv1D(
            filters=config['conv1d_2']['filter'],
            kernel_size=config['conv1d_2']['kernel'],
            activation=config['conv1d_2']['activation'],
            padding=config['conv1d_2']['padding']
        )
        self.glob_maxpool1d = tkl.GlobalAveragePooling1D()
        self.dropout = tkl.Dropout(config['dropout'])
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))
        self.conv3 = tkl.Conv1D(n_out_ft, 1, padding='same')

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.bilstm1(x)
        x = self.attention1(x)
        x = self.conv1(x)
        x = self.maxpool1d(x)
        x = self.bilstm2(x)
        x = self.attention2(x)
        x = self.conv2(x)
        x = self.glob_maxpool1d(x)
        x = self.dropout(x)
        x = self.dense1(x)
        x = self.dense2(x)
        x = self.conv3(x)
        return x


class bilstm_multihead_attention_conv1d_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(bilstm_multihead_attention_conv1d_dense, self).__init__()
        if config is None:
            config = {
                'bilstm1': {'units': 64},
                'conv1d_1': {
                    'filter': 128,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'same'
                },
                'dropout': 0.5,
                'bilstm2': {'units': 64},
                'conv1d_2': {
                    'filter': 256,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'same'
                },
                'dense': {'activation': 'relu'}
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.bilstm1 = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm1']['units'],
                return_sequences=True
            )
        )
        self.multihead_attention = tkl.MultiHeadAttention(num_heads=2, key_dim=2, dropout=0.1)
        self.conv1 = tkl.Conv1D(
            filters=config['conv1d_1']['filter'],
            kernel_size=config['conv1d_1']['kernel'],
            activation=config['conv1d_1']['activation'],
            padding=config['conv1d_1']['padding']
        )
        self.maxpool1d = tkl.MaxPool1D()
        self.bilstm2 = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm2']['units'],
                return_sequences=True
            )
        )
        self.conv2 = tkl.Conv1D(
            filters=config['conv1d_2']['filter'],
            kernel_size=config['conv1d_2']['kernel'],
            activation=config['conv1d_2']['activation'],
            padding=config['conv1d_2']['padding']
        )
        self.glob_maxpool1d = tkl.GlobalAveragePooling1D()
        self.dropout = tkl.Dropout(config['dropout'])
        self.dense = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.reshape = tkl.Reshape((n_future, n_out_ft))
        self.conv3 = tkl.Conv1D(filters=n_out_ft, kernel_size=1, padding='same')

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.bilstm1(x)
        x = self.multihead_attention(x, x)
        x = self.conv1(x)
        x = self.maxpool1d(x)
        x = self.bilstm2(x)
        x = self.multihead_attention(x, x)
        x = self.conv2(x)
        x = self.glob_maxpool1d(x)
        x = self.dropout(x)
        x = self.dense(x)
        x = self.reshape(x)
        x = self.conv3(x)
        return x


class bilstm_luong_attention_conv1d_dense(tkm.Model):

    def __init__(
            self,
            n_future=None,
            n_out_ft=None,
            config=None
    ):
        super(bilstm_luong_attention_conv1d_dense, self).__init__()
        if config is None:
            config = {
                'bilstm1': {'units': 64},
                'conv1d_1': {
                    'filter': 128,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'same'
                },
                'dropout': 0.5,
                'bilstm2': {'units': 32},
                'conv1d_2': {
                    'filter': 256,
                    'kernel': 3,
                    'activation': 'relu',
                    'padding': 'same'
                },
                'dense': {'activation': 'relu'}
            }
        self.mask = tkl.Masking(mask_value=value_miss)
        self.bilstm1 = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm1']['units'],
                return_sequences=True
            )
        )
        self.attention1 = SeqSelfAttention(
            attention_type=SeqSelfAttention.ATTENTION_TYPE_MUL,
            kernel_regularizer=tkr.l2(1e-4),
            bias_regularizer=tkr.l1(1e-4),
            attention_regularizer_weight=1e-4,
            name='Attention1'
        )
        self.conv1 = tkl.Conv1D(
            filters=config['conv1d_1']['filter'],
            kernel_size=config['conv1d_1']['kernel'],
            activation=config['conv1d_1']['activation'],
            padding=config['conv1d_1']['padding']
        )
        self.maxpool1d = tkl.MaxPool1D()
        self.bilstm2 = tkl.Bidirectional(
            tkl.LSTM(
                units=config['bilstm2']['units'],
                return_sequences=True
            )
        )
        self.attention2 = SeqSelfAttention(
            attention_type=SeqSelfAttention.ATTENTION_TYPE_MUL,
            kernel_regularizer=tkr.l2(1e-4),
            bias_regularizer=tkr.l1(1e-4),
            attention_regularizer_weight=1e-4,
            name='Attention2'
        )
        self.conv2 = tkl.Conv1D(
            filters=config['conv1d_2']['filter'],
            kernel_size=config['conv1d_2']['kernel'],
            activation=config['conv1d_2']['activation'],
            padding=config['conv1d_2']['padding']
        )
        self.glob_maxpool1d = tkl.GlobalAveragePooling1D()
        self.dropout = tkl.Dropout(config['dropout'])
        self.dense1 = tkl.Dense(
            n_future * n_out_ft,
            activation=config['dense']['activation']
        )
        self.dense2 = tkl.Reshape((n_future, n_out_ft))
        self.conv3 = tkl.Conv1D(n_out_ft, 1, padding='same')

    def call(self, inputs, training=True, **kwargs):
        x = self.mask(inputs)
        x = self.bilstm1(x)
        x = self.attention1(x)
        x = self.conv1(x)
        x = self.maxpool1d(x)
        x = self.bilstm2(x)
        x = self.attention2(x)
        x = self.conv2(x)
        x = self.glob_maxpool1d(x)
        x = self.dropout(x)
        x = self.dense1(x)
        x = self.dense2(x)
        x = self.conv3(x)
        return x